from datetime import datetime
from unittest import TestCase

import boto3
from botocore.exceptions import ClientError
from pytest import raises

from moto import mock_aws


@mock_aws
class TestAutoScalingScheduledActions(TestCase):
    def setUp(self) -> None:
        self.client = boto3.client("autoscaling", region_name="us-east-1")
        self.asg_name = "tester_group"

    def test_list_many_scheduled_scaling_actions(self):
        for i in range(30):
            self._create_scheduled_action(name=f"my-scheduled-action-{i}", idx=i)

        response = self.client.describe_scheduled_actions(
            AutoScalingGroupName=self.asg_name
        )
        actions = response["ScheduledUpdateGroupActions"]
        assert len(actions) == 30

    def test_non_existing_group_name(self):
        self._create_scheduled_action(name="my-scheduled-action", idx=1)

        response = self.client.describe_scheduled_actions(
            AutoScalingGroupName="wrong_group"
        )
        actions = response["ScheduledUpdateGroupActions"]
        # since there is no such group name, no actions have been returned
        assert len(actions) == 0

    def test_describe_scheduled_actions_returns_all_actions_when_no_argument_is_passed(
        self,
    ):
        for i in range(30):
            self._create_scheduled_action(name=f"my-scheduled-action-{i}", idx=i)

        for i in range(10):
            self._create_scheduled_action(
                name=f"my-scheduled-action-4{i}", idx=i, asg_name="test_group-2"
            )

        response = self.client.describe_scheduled_actions()
        actions = response["ScheduledUpdateGroupActions"]

        # Since no argument is passed describe_scheduled_actions, all scheduled actions are returned
        assert len(actions) == 40

    def test_scheduled_action_delete(self):
        for i in range(3):
            self._create_scheduled_action(name=f"my-scheduled-action-{i}", idx=i)

        response = self.client.describe_scheduled_actions(
            AutoScalingGroupName=self.asg_name
        )
        actions = response["ScheduledUpdateGroupActions"]
        assert len(actions) == 3

        self.client.delete_scheduled_action(
            AutoScalingGroupName=self.asg_name,
            ScheduledActionName="my-scheduled-action-2",
        )
        self.client.delete_scheduled_action(
            AutoScalingGroupName=self.asg_name,
            ScheduledActionName="my-scheduled-action-1",
        )
        response = self.client.describe_scheduled_actions(
            AutoScalingGroupName=self.asg_name
        )
        actions = response["ScheduledUpdateGroupActions"]
        assert len(actions) == 1

    def test_delete_nonexistent_action(self) -> None:
        with raises(ClientError) as exc_info:
            self.client.delete_scheduled_action(
                AutoScalingGroupName=self.asg_name,
                ScheduledActionName="nonexistent",
            )
        assert exc_info.value.response["Error"]["Code"] == "ValidationError"
        assert (
            exc_info.value.response["Error"]["Message"]
            == "Scheduled action name not found"
        )

    def test_put_actions_content(self) -> None:
        # actions with different combinations of properties return the correct content
        self.client.batch_put_scheduled_update_group_action(
            AutoScalingGroupName=self.asg_name,
            ScheduledUpdateGroupActions=[
                {
                    "ScheduledActionName": "desired-capacity-with-start-time",
                    "DesiredCapacity": 1,
                    "StartTime": "2024-01-18T09:00:00Z",
                },
                {
                    "ScheduledActionName": "min-size-with-recurrence",
                    "MinSize": 3,
                    "Recurrence": "* * * * *",
                },
                {
                    "ScheduledActionName": "complete",
                    "MinSize": 1,
                    "DesiredCapacity": 2,
                    "MaxSize": 3,
                    "Recurrence": "* * * * *",
                    "StartTime": "2024-01-19T09:00:00Z",
                    "EndTime": "2024-01-20T09:00:00Z",
                    "TimeZone": "America/New_York",
                },
            ],
        )
        response = self.client.describe_scheduled_actions(
            AutoScalingGroupName=self.asg_name
        )
        actions = response["ScheduledUpdateGroupActions"]
        assert len(actions) == 3

        desired_capacity_action = next(
            filter(
                lambda action: action["ScheduledActionName"]
                == "desired-capacity-with-start-time",
                actions,
            )
        )
        assert desired_capacity_action["DesiredCapacity"] == 1
        assert desired_capacity_action["StartTime"] == datetime.fromisoformat(
            "2024-01-18T09:00:00+00:00"
        )
        assert "MinSize" not in desired_capacity_action
        assert "MaxSize" not in desired_capacity_action
        assert "TimeZone" not in desired_capacity_action
        assert "Recurrence" not in desired_capacity_action
        assert "EndTime" not in desired_capacity_action

        min_size_action = next(
            filter(
                lambda action: action["ScheduledActionName"]
                == "min-size-with-recurrence",
                actions,
            )
        )
        assert min_size_action["MinSize"] == 3
        assert min_size_action["Recurrence"] == "* * * * *"
        assert "DesiredCapacity" not in min_size_action
        assert "MaxSize" not in min_size_action
        assert "TimeZone" not in min_size_action
        # StartTime may be present
        assert "EndTime" not in min_size_action

        complete_action = next(
            filter(lambda action: action["ScheduledActionName"] == "complete", actions)
        )
        assert complete_action["MinSize"] == 1
        assert complete_action["DesiredCapacity"] == 2
        assert complete_action["MaxSize"] == 3
        assert complete_action["Recurrence"] == "* * * * *"
        assert complete_action["StartTime"] == datetime.fromisoformat(
            "2024-01-19T09:00:00+00:00"
        )
        assert complete_action["EndTime"] == datetime.fromisoformat(
            "2024-01-20T09:00:00+00:00"
        )
        assert complete_action["TimeZone"] == "America/New_York"

    def test_put_replaces_action_with_same_name(self) -> None:
        self.client.put_scheduled_update_group_action(
            AutoScalingGroupName=self.asg_name,
            ScheduledActionName="my-action",
            Recurrence="* * * * *",
            MinSize=3,
        )
        self.client.put_scheduled_update_group_action(
            AutoScalingGroupName=self.asg_name,
            ScheduledActionName="my-action",
            Recurrence="* * * * *",
            DesiredCapacity=2,
        )

        response = self.client.describe_scheduled_actions(
            AutoScalingGroupName=self.asg_name
        )
        actions = response["ScheduledUpdateGroupActions"]
        assert len(actions) == 1

        assert actions[0]["DesiredCapacity"] == 2
        assert "MinSize" not in actions[0]

    def _create_scheduled_action(self, name, idx, asg_name=None):
        self.client.put_scheduled_update_group_action(
            AutoScalingGroupName=asg_name or self.asg_name,
            ScheduledActionName=name,
            StartTime=f"2022-07-01T00:00:{idx}Z",
            EndTime=f"2022-09-01T00:00:{idx}Z",
            Recurrence="* * * * *",
            MinSize=idx + 1,
            MaxSize=idx + 5,
            DesiredCapacity=idx + 3,
            TimeZone="Etc/UTC",
        )

    def test_batch_put_scheduled_group_action(self) -> None:
        put_response = self.client.batch_put_scheduled_update_group_action(
            AutoScalingGroupName=self.asg_name,
            ScheduledUpdateGroupActions=[
                {
                    "ScheduledActionName": "StartAction",
                    "Recurrence": "0 9 * * 1-5",
                    "MinSize": 1,
                    "MaxSize": 1,
                    "DesiredCapacity": 1,
                },
                {
                    "ScheduledActionName": "StopAction",
                    "Recurrence": "0 17 * * 1-5",
                    "MinSize": 0,
                    "MaxSize": 0,
                    "DesiredCapacity": 0,
                },
            ],
        )
        assert len(put_response["FailedScheduledUpdateGroupActions"]) == 0

        response = self.client.describe_scheduled_actions(
            AutoScalingGroupName=self.asg_name
        )
        actions = response["ScheduledUpdateGroupActions"]
        assert len(actions) == 2

    def test_batch_delete_scheduled_action(self) -> None:
        action_names = [f"my-action-{i}" for i in range(10)]
        for j, action_name in enumerate(action_names):
            self._create_scheduled_action(action_name, j)
        delete_response = self.client.batch_delete_scheduled_action(
            AutoScalingGroupName=self.asg_name,
            ScheduledActionNames=action_names[::2],  # delete even indices
        )
        assert len(delete_response["FailedScheduledActions"]) == 0

        describe_response = self.client.describe_scheduled_actions(
            AutoScalingGroupName=self.asg_name
        )
        actions = describe_response["ScheduledUpdateGroupActions"]
        assert len(actions) == 5

        delete_response = self.client.batch_delete_scheduled_action(
            AutoScalingGroupName=self.asg_name,
            ScheduledActionNames=action_names[1::2],  # delete odd indices
        )
        assert len(delete_response["FailedScheduledActions"]) == 0

        describe_response = self.client.describe_scheduled_actions(
            AutoScalingGroupName=self.asg_name
        )
        actions = describe_response["ScheduledUpdateGroupActions"]
        assert len(actions) == 0

    def test_batch_delete_nonexistent_action(self) -> None:
        action_name = "nonexistent"
        response = self.client.batch_delete_scheduled_action(
            AutoScalingGroupName=self.asg_name,
            ScheduledActionNames=[action_name],
        )
        assert len(response["FailedScheduledActions"]) == 1
        assert response["FailedScheduledActions"][0] == {
            "ScheduledActionName": action_name,
            "ErrorCode": "ValidationError",
            "ErrorMessage": "Scheduled action name not found",
        }



"""Exceptions raised by the dsql service."""

from moto.core.exceptions import JsonRESTError


class ValidationException(JsonRESTError):
    """Tag validation failed."""

    code = 400

    def __init__(self, message: str):
        super().__init__("ValidationException", message)



from __future__ import annotations

import abc
import datetime
import logging
from abc import ABC
from typing import Any, Optional

from moto.stepfunctions.parser.api import (
    ExecutionFailedEventDetails,
    HistoryEventExecutionDataDetails,
    HistoryEventType,
    StateEnteredEventDetails,
    StateExitedEventDetails,
)
from moto.stepfunctions.parser.asl.component.common.assign.assign_decl import AssignDecl
from moto.stepfunctions.parser.asl.component.common.catch.catch_outcome import (
    CatchOutcome,
)
from moto.stepfunctions.parser.asl.component.common.comment import Comment
from moto.stepfunctions.parser.asl.component.common.error_name.failure_event import (
    FailureEvent,
    FailureEventException,
)
from moto.stepfunctions.parser.asl.component.common.error_name.states_error_name import (
    StatesErrorName,
)
from moto.stepfunctions.parser.asl.component.common.error_name.states_error_name_type import (
    StatesErrorNameType,
)
from moto.stepfunctions.parser.asl.component.common.flow.end import End
from moto.stepfunctions.parser.asl.component.common.flow.next import Next
from moto.stepfunctions.parser.asl.component.common.outputdecl import Output
from moto.stepfunctions.parser.asl.component.common.path.input_path import (
    InputPath,
    InputPathBase,
)
from moto.stepfunctions.parser.asl.component.common.path.output_path import (
    OutputPath,
    OutputPathBase,
)
from moto.stepfunctions.parser.asl.component.common.query_language import (
    QueryLanguage,
    QueryLanguageMode,
)
from moto.stepfunctions.parser.asl.component.eval_component import EvalComponent
from moto.stepfunctions.parser.asl.component.state.state_continue_with import (
    ContinueWith,
    ContinueWithEnd,
    ContinueWithNext,
)
from moto.stepfunctions.parser.asl.component.state.state_props import StateProps
from moto.stepfunctions.parser.asl.component.state.state_type import StateType
from moto.stepfunctions.parser.asl.eval.environment import Environment
from moto.stepfunctions.parser.asl.eval.event.event_detail import EventDetails
from moto.stepfunctions.parser.asl.eval.program_state import ProgramRunning
from moto.stepfunctions.parser.asl.eval.states import StateData
from moto.stepfunctions.parser.asl.utils.encoding import to_json_str
from moto.stepfunctions.parser.quotas import is_within_size_quota

LOG = logging.getLogger(__name__)


class CommonStateField(EvalComponent, ABC):
    name: str

    query_language: QueryLanguage

    # The state's type.
    state_type: StateType

    # There can be any number of terminal states per state machine. Only one of Next or End can
    # be used in a state. Some state types, such as Choice, don't support or use the End field.
    continue_with: ContinueWith

    # Holds a human-readable description of the state.
    comment: Optional[Comment]

    # A path that selects a portion of the state's input to be passed to the state's state_task for processing.
    # If omitted, it has the value $ which designates the entire input.
    input_path: Optional[InputPath]

    # A path that selects a portion of the state's output to be passed to the next state.
    # If omitted, it has the value $ which designates the entire output.
    output_path: Optional[OutputPath]

    assign_decl: Optional[AssignDecl]

    output: Optional[Output]

    state_entered_event_type: HistoryEventType
    state_exited_event_type: Optional[HistoryEventType]

    def __init__(
        self,
        state_entered_event_type: HistoryEventType,
        state_exited_event_type: Optional[HistoryEventType],
    ):
        self.comment = None
        self.input_path = InputPathBase(InputPathBase.DEFAULT_PATH)
        self.output_path = OutputPathBase(OutputPathBase.DEFAULT_PATH)
        self.state_entered_event_type = state_entered_event_type
        self.state_exited_event_type = state_exited_event_type

    def from_state_props(self, state_props: StateProps) -> None:
        self.name = state_props.name
        self.query_language = state_props.get(QueryLanguage) or QueryLanguage()
        self.state_type = state_props.get(StateType)
        self.continue_with = (
            ContinueWithEnd()
            if state_props.get(End)
            else ContinueWithNext(state_props.get(Next))
        )
        self.comment = state_props.get(Comment)
        self.assign_decl = state_props.get(AssignDecl)
        # JSONPath sub-productions.
        if self.query_language.query_language_mode == QueryLanguageMode.JSONPath:
            self.input_path = state_props.get(InputPath) or InputPathBase(
                InputPathBase.DEFAULT_PATH
            )
            self.output_path = state_props.get(OutputPath) or OutputPathBase(
                OutputPathBase.DEFAULT_PATH
            )
            self.output = None
        # JSONata sub-productions.
        else:
            self.input_path = None
            self.output_path = None
            self.output = state_props.get(Output)

    def _set_next(self, env: Environment) -> None:
        if env.next_state_name != self.name:
            # Next was already overridden.
            return

        if isinstance(self.continue_with, ContinueWithNext):
            env.next_state_name = self.continue_with.next_state.name
        elif isinstance(
            self.continue_with, ContinueWithEnd
        ):  # This includes ContinueWithSuccess
            env.set_ended()
        else:
            LOG.error(
                "Could not handle ContinueWith type of '%s'.", type(self.continue_with)
            )

    def _is_language_query_jsonpath(self) -> bool:
        return self.query_language.query_language_mode == QueryLanguageMode.JSONPath

    def _get_state_entered_event_details(
        self, env: Environment
    ) -> StateEnteredEventDetails:
        return StateEnteredEventDetails(
            name=self.name,
            input=to_json_str(env.states.get_input(), separators=(",", ":")),
            inputDetails=HistoryEventExecutionDataDetails(
                truncated=False  # Always False for api calls.
            ),
        )

    def _get_state_exited_event_details(
        self, env: Environment
    ) -> StateExitedEventDetails:
        event_details = StateExitedEventDetails(
            name=self.name,
            output=to_json_str(env.states.get_input(), separators=(",", ":")),
            outputDetails=HistoryEventExecutionDataDetails(
                truncated=False  # Always False for api calls.
            ),
        )
        # TODO add typing when these become available in boto.
        assigned_variables = env.variable_store.get_assigned_variables()
        env.variable_store.reset_tracing()
        if assigned_variables:
            event_details["assignedVariables"] = assigned_variables  # noqa
            event_details["assignedVariablesDetails"] = {"truncated": False}  # noqa
        return event_details

    def _verify_size_quota(self, env: Environment, value: Any) -> None:
        is_within: bool = is_within_size_quota(value)
        if is_within:
            return
        error_type = StatesErrorNameType.StatesStatesDataLimitExceeded
        cause = (
            f"The state/task '{self.name}' returned a result with a size exceeding "
            f"the maximum number of bytes service limit."
        )
        raise FailureEventException(
            failure_event=FailureEvent(
                env=env,
                error_name=StatesErrorName(typ=error_type),
                event_type=HistoryEventType.TaskFailed,
                event_details=EventDetails(
                    executionFailedEventDetails=ExecutionFailedEventDetails(
                        error=error_type.to_name(),
                        cause=cause,
                    )
                ),
            )
        )

    @abc.abstractmethod
    def _eval_state(self, env: Environment) -> None: ...

    def _eval_body(self, env: Environment) -> None:
        env.event_manager.add_event(
            context=env.event_history_context,
            event_type=self.state_entered_event_type,
            event_details=EventDetails(
                stateEnteredEventDetails=self._get_state_entered_event_details(env=env)
            ),
        )

        env.states.context_object.context_object_data["State"] = StateData(
            EnteredTime=datetime.datetime.now(tz=datetime.timezone.utc).isoformat(),
            Name=self.name,
        )

        # Filter the input onto the stack.
        if self.input_path:
            self.input_path.eval(env)
        else:
            env.stack.append(env.states.get_input())

        # Exec the state's logic.
        self._eval_state(env)
        #
        if not isinstance(env.program_state(), ProgramRunning):
            return

        # Obtain a reference to the state output.
        output = env.stack[-1]

        # CatcherOutputs (i.e. outputs of Catch blocks) are never subjects of output normalisers,
        # the entire value is instead passed by value as input to the next state, or program output.
        if not isinstance(output, CatchOutcome):
            # Ensure the state's output is within state size quotas.
            self._verify_size_quota(env=env, value=output)

            # Process output value as next state input.
            if self.output_path:
                self.output_path.eval(env=env)
            elif self.output:
                self.output.eval(env=env)
            else:
                current_output = env.stack.pop()
                env.states.reset(input_value=current_output)

            # Set next state or halt (end).
            self._set_next(env)

        if self.state_exited_event_type is not None:
            env.event_manager.add_event(
                context=env.event_history_context,
                event_type=self.state_exited_event_type,
                event_details=EventDetails(
                    stateExitedEventDetails=self._get_state_exited_event_details(
                        env=env
                    ),
                ),
            )



"""Unit tests for kafka-supported APIs."""

import boto3

from moto import mock_aws

# See our Development Tips on writing tests for hints on how to write good tests:
# http://docs.getmoto.org/en/latest/docs/contributing/development_tips/tests.html


FAKE_TAGS = {"TestKey": "TestValue", "TestKey2": "TestValue2"}


@mock_aws
def test_create_cluster_v2():
    client = boto3.client("kafka", region_name="ap-southeast-1")
    s_cluster_name = "TestServerlessCluster"
    p_cluster_name = "TestProvisionedCluster"

    s_response = client.create_cluster_v2(
        ClusterName=s_cluster_name,
        Serverless={
            "VpcConfigs": [
                {
                    "SubnetIds": ["subnet-0123456789abcdef0"],
                    "SecurityGroupIds": ["sg-0123456789abcdef0"],
                }
            ]
        },
        Tags=FAKE_TAGS,
    )

    p_response = client.create_cluster_v2(
        ClusterName=p_cluster_name,
        Provisioned={
            "BrokerNodeGroupInfo": {
                "InstanceType": "kafka.m5.large",
                "ClientSubnets": ["subnet-0123456789abcdef0"],
                "SecurityGroups": ["sg-0123456789abcdef0"],
            },
            "KafkaVersion": "2.8.1",
            "NumberOfBrokerNodes": 3,
        },
        Tags=FAKE_TAGS,
    )

    assert s_response["ClusterArn"].startswith("arn:aws:kafka")
    assert s_response["ClusterName"] == s_cluster_name
    assert s_response["State"] == "CREATING"

    assert p_response["ClusterArn"].startswith("arn:aws:kafka")
    assert p_response["ClusterName"] == p_cluster_name
    assert p_response["State"] == "CREATING"

    clusters = client.list_clusters_v2()
    assert len(clusters["ClusterInfoList"]) == 2
    assert clusters["ClusterInfoList"][0]["ClusterName"] == s_cluster_name
    assert clusters["ClusterInfoList"][0]["ClusterType"] == "SERVERLESS"
    assert clusters["ClusterInfoList"][1]["ClusterName"] == p_cluster_name
    assert clusters["ClusterInfoList"][1]["ClusterType"] == "PROVISIONED"

    s_resp = client.describe_cluster_v2(ClusterArn=s_response["ClusterArn"])
    s_cluster_info = s_resp["ClusterInfo"]
    p_resp = client.describe_cluster_v2(ClusterArn=p_response["ClusterArn"])
    p_cluster_info = p_resp["ClusterInfo"]

    assert s_cluster_info["ClusterName"] == s_cluster_name
    assert s_cluster_info["State"] == "CREATING"
    assert s_cluster_info["ClusterType"] == "SERVERLESS"
    assert s_cluster_info["Serverless"]["VpcConfigs"][0]["SubnetIds"] == [
        "subnet-0123456789abcdef0"
    ]
    assert s_cluster_info["Serverless"]["VpcConfigs"][0]["SecurityGroupIds"] == [
        "sg-0123456789abcdef0"
    ]
    assert s_cluster_info["Tags"] == FAKE_TAGS

    assert p_cluster_info["ClusterName"] == p_cluster_name
    assert p_cluster_info["State"] == "CREATING"
    assert p_cluster_info["ClusterType"] == "PROVISIONED"
    assert (
        p_cluster_info["Provisioned"]["BrokerNodeGroupInfo"]["InstanceType"]
        == "kafka.m5.large"
    )


@mock_aws
def test_list_tags_for_resource():
    client = boto3.client("kafka", region_name="us-east-2")
    create_resp = client.create_cluster(
        ClusterName="TestCluster",
        BrokerNodeGroupInfo={
            "InstanceType": "kafka.m5.large",
            "ClientSubnets": ["subnet-0123456789abcdef0"],
            "SecurityGroups": ["sg-0123456789abcdef0"],
        },
        KafkaVersion="2.8.1",
        NumberOfBrokerNodes=3,
        Tags=FAKE_TAGS,
    )

    TempTags = {"TestKey3": "TestValue3"}

    client.tag_resource(
        ResourceArn=create_resp["ClusterArn"],
        Tags=TempTags,
    )

    tags = client.list_tags_for_resource(ResourceArn=create_resp["ClusterArn"])
    assert tags["Tags"] == {**FAKE_TAGS, **TempTags}

    client.untag_resource(
        ResourceArn=create_resp["ClusterArn"],
        TagKeys=["TestKey3"],
    )

    tags = client.list_tags_for_resource(ResourceArn=create_resp["ClusterArn"])

    assert tags["Tags"] == FAKE_TAGS


@mock_aws
def test_create_cluster():
    client = boto3.client("kafka", region_name="eu-west-1")
    cluster_name = "TestCluster"
    response = client.create_cluster(
        ClusterName=cluster_name,
        BrokerNodeGroupInfo={
            "InstanceType": "kafka.m5.large",
            "ClientSubnets": ["subnet-0123456789abcdef0"],
            "SecurityGroups": ["sg-0123456789abcdef0"],
        },
        KafkaVersion="2.8.1",
        NumberOfBrokerNodes=3,
        Tags=FAKE_TAGS,
    )

    assert response["ClusterArn"].startswith("arn:aws:kafka")
    assert response["ClusterName"] == cluster_name
    assert response["State"] == "CREATING"

    clusters = client.list_clusters()
    assert len(clusters["ClusterInfoList"]) == 1
    assert clusters["ClusterInfoList"][0]["ClusterName"] == cluster_name

    resp = client.describe_cluster(ClusterArn=response["ClusterArn"])
    assert resp["ClusterInfo"]["ClusterName"] == cluster_name
    assert resp["ClusterInfo"]["State"] == "CREATING"
    assert resp["ClusterInfo"]["CurrentBrokerSoftwareInfo"]["KafkaVersion"] == "2.8.1"
    assert resp["ClusterInfo"]["NumberOfBrokerNodes"] == 3
    assert (
        resp["ClusterInfo"]["BrokerNodeGroupInfo"]["InstanceType"] == "kafka.m5.large"
    )
    assert resp["ClusterInfo"]["BrokerNodeGroupInfo"]["ClientSubnets"] == [
        "subnet-0123456789abcdef0"
    ]
    assert resp["ClusterInfo"]["BrokerNodeGroupInfo"]["SecurityGroups"] == [
        "sg-0123456789abcdef0"
    ]
    assert resp["ClusterInfo"]["Tags"] == FAKE_TAGS


@mock_aws
def test_delete_cluster():
    client = boto3.client("kafka", region_name="us-east-2")
    create_resp = client.create_cluster(
        ClusterName="TestCluster",
        BrokerNodeGroupInfo={
            "InstanceType": "kafka.m5.large",
            "ClientSubnets": ["subnet-0123456789abcdef0"],
            "SecurityGroups": ["sg-0123456789abcdef0"],
        },
        KafkaVersion="2.8.1",
        NumberOfBrokerNodes=3,
        Tags=FAKE_TAGS,
    )

    client.delete_cluster(ClusterArn=create_resp["ClusterArn"])
    clusters = client.list_clusters()
    assert len(clusters["ClusterInfoList"]) == 0


@mock_aws
def test_list_clusters_v2():
    client = boto3.client("kafka", region_name="ap-southeast-1")
    s_cluster_name = "TestServerlessCluster"
    p_cluster_name = "TestProvisionedCluster"

    client.create_cluster_v2(
        ClusterName=s_cluster_name,
        Serverless={
            "VpcConfigs": [
                {
                    "SubnetIds": ["subnet-0123456789abcdef0"],
                    "SecurityGroupIds": ["sg-0123456789abcdef0"],
                }
            ]
        },
        Tags=FAKE_TAGS,
    )

    client.create_cluster_v2(
        ClusterName=p_cluster_name,
        Provisioned={
            "BrokerNodeGroupInfo": {
                "InstanceType": "kafka.m5.large",
                "ClientSubnets": ["subnet-0123456789abcdef0"],
                "SecurityGroups": ["sg-0123456789abcdef0"],
            },
            "KafkaVersion": "2.8.1",
            "NumberOfBrokerNodes": 3,
        },
        Tags=FAKE_TAGS,
    )

    clusters = client.list_clusters_v2()

    assert len(clusters["ClusterInfoList"]) == 2
    for cluster in clusters["ClusterInfoList"]:
        assert "ActiveOperationArn" in cluster
        assert "ClusterType" in cluster
        assert "ClusterArn" in cluster
        assert "ClusterName" in cluster
        assert "CreationTime" in cluster
        assert "CurrentVersion" in cluster
        assert "State" in cluster
        assert "StateInfo" in cluster
        assert "Tags" in cluster

        cluster_type = cluster["ClusterType"]
        assert cluster_type[0].upper() + cluster_type[1:].lower() in cluster



import abc
import logging
import threading
from typing import Any, Callable, Dict, List, TypedDict, Union

from moto.moto_api._internal import mock_random

log = logging.getLogger(__name__)

ExistingIds = Union[List[str], None]
Tags = Union[Dict[str, str], List[Dict[str, str]], None]

# Custom resource tag to override the generated resource ID.
TAG_KEY_CUSTOM_ID = "_custom_id_"


class IdSourceContext(TypedDict, total=False):
    resource_identifier: "ResourceIdentifier"
    tags: Tags
    existing_ids: ExistingIds


class ResourceIdentifier(abc.ABC):
    """
    Base class for resource identifiers. When implementing a new resource, it is important to set
    the service and resource as they will be used to create the unique identifier for that resource.

    It is recommended to implement the `generate` method using functions decorated with `@moto_id`.
    This will ensure that your resource can be assigned a custom id.
    """

    service: str
    resource: str

    def __init__(self, account_id: str, region: str, name: str):
        self.account_id = account_id
        self.region = region
        self.name = name or ""

    @abc.abstractmethod
    def generate(self, existing_ids: ExistingIds = None, tags: Tags = None) -> str:
        """Method to generate a resource id"""

    @property
    def unique_identifier(self) -> str:
        return ".".join(
            [self.account_id, self.region, self.service, self.resource, self.name]
        )

    def __str__(self) -> str:
        return self.unique_identifier


class MotoIdManager:
    """class to manage custom ids. Do not create instance and instead
    use the `id_manager` instance created below."""

    _custom_ids: Dict[str, str]
    _id_sources: List[Callable[[IdSourceContext], Union[str, None]]]

    _lock: threading.RLock

    def __init__(self) -> None:
        self._custom_ids = {}
        self._lock = threading.RLock()
        self._id_sources = []

        self.add_id_source(self.get_id_from_tags)
        self.add_id_source(self.get_custom_id_from_context)

    def get_custom_id(
        self, resource_identifier: ResourceIdentifier
    ) -> Union[str, None]:
        # retrieves a custom_id for a resource. Returns None if no id were registered
        # that matches the `resource_identifier`
        return self._custom_ids.get(resource_identifier.unique_identifier)

    def set_custom_id(
        self, resource_identifier: ResourceIdentifier, custom_id: str
    ) -> None:
        # Do not set a custom_id for a resource no value was found for the name
        if not resource_identifier.name:
            return
        with self._lock:
            self._custom_ids[resource_identifier.unique_identifier] = custom_id

    def unset_custom_id(self, resource_identifier: ResourceIdentifier) -> None:
        # removes a set custom_id for a resource
        with self._lock:
            self._custom_ids.pop(resource_identifier.unique_identifier, None)

    def add_id_source(
        self, id_source: Callable[[IdSourceContext], Union[str, None]]
    ) -> None:
        self._id_sources.append(id_source)

    @staticmethod
    def get_id_from_tags(id_source_context: IdSourceContext) -> Union[str, None]:
        if not (tags := id_source_context.get("tags")):
            return None

        if isinstance(tags, dict):
            return tags.get(TAG_KEY_CUSTOM_ID)

        if isinstance(tags, list):
            return next(
                (
                    tag.get("Value")
                    for tag in tags
                    if tag.get("Key") == TAG_KEY_CUSTOM_ID
                ),
                None,
            )

    def get_custom_id_from_context(
        self, id_source_context: IdSourceContext
    ) -> Union[str, None]:
        # retrieves a custom_id for a resource. Returns None
        if resource_identifier := id_source_context.get("resource_identifier"):
            return self.get_custom_id(resource_identifier)
        return None

    def find_id_from_sources(
        self, id_source_context: IdSourceContext
    ) -> Union[str, None]:
        existing_ids = id_source_context.get("existing_ids") or []
        for id_source in self._id_sources:
            if found_id := id_source(id_source_context):
                if found_id in existing_ids:
                    log.debug(
                        f"Found id {found_id} for resource {id_source_context.get('resource_identifier')}, "
                        "but a resource already exists with this id."
                    )
                else:
                    return found_id

        return None


moto_id_manager = MotoIdManager()


def moto_id(fn: Callable[..., str]) -> Callable[..., str]:
    """
    Decorator for helping in creation of static ids.

    The decorated function should accept the following parameters

    :param resource_identifier
    :param existing_ids
        If provided, we will omit returning a custom id if it is already on the list
    :param tags
        If provided will look for a tag named `_custom_id_`. This will take precedence over registered custom ids
    """

    def _wrapper(
        resource_identifier: ResourceIdentifier,
        existing_ids: ExistingIds = None,
        tags: Tags = None,
        **kwargs: Dict[str, Any],
    ) -> str:
        if resource_identifier and (
            found_id := moto_id_manager.find_id_from_sources(
                IdSourceContext(
                    resource_identifier=resource_identifier,
                    existing_ids=existing_ids,
                    tags=tags,
                )
            )
        ):
            return found_id

        return fn(
            resource_identifier=resource_identifier,
            existing_ids=existing_ids,
            tags=tags,
            **kwargs,
        )

    return _wrapper


@moto_id
def generate_str_id(  # type: ignore
    resource_identifier: ResourceIdentifier,
    existing_ids: ExistingIds = None,
    tags: Tags = None,
    length: int = 20,
    include_digits: bool = True,
    lower_case: bool = False,
) -> str:
    return mock_random.get_random_string(length, include_digits, lower_case)



"""AuroraDSQLBackend class with methods for supported APIs."""

from collections import OrderedDict
from typing import Any, Dict, Optional

from moto.core.base_backend import BackendDict, BaseBackend
from moto.core.common_models import BaseModel
from moto.core.utils import iso_8601_datetime_with_milliseconds, utcnow
from moto.moto_api._internal import mock_random
from moto.moto_api._internal.managed_state_model import ManagedState
from moto.utilities.utils import get_partition

from .exceptions import ValidationException


class Cluster(BaseModel, ManagedState):
    """Model for an AuroraDSQL cluster."""

    def to_dict(self) -> Dict[str, Any]:
        dct = {
            "identifier": self.identifier,
            "arn": self.arn,
            "status": self.status,
            "creationTime": iso_8601_datetime_with_milliseconds(self.creation_time),
            "deletionProtectionEnabled": self.deletion_protection_enabled,
        }
        return {k: v for k, v in dct.items() if v is not None}

    def __init__(
        self,
        region_name: str,
        account_id: str,
        deletion_protection_enabled: Optional[bool],
        tags: Optional[Dict[str, str]],
        client_token: Optional[str],
    ):
        ManagedState.__init__(
            self, "dsql::cluster", transitions=[("CREATING", "ACTIVE")]
        )
        self.region_name = region_name
        self.account_id = account_id
        self.identifier = mock_random.get_random_hex(26)
        self.arn = f"arn:{get_partition(self.region_name)}:dsql:{self.region_name}:{self.account_id}:cluster/{self.identifier}"
        self.creation_time = utcnow()
        self.deletion_protection_enabled = deletion_protection_enabled
        self.tags = tags
        self.client_token = client_token


class AuroraDSQLBackend(BaseBackend):
    """Implementation of AuroraDSQL APIs."""

    def __init__(self, region_name: str, account_id: str):
        super().__init__(region_name, account_id)
        self.region_name = region_name
        self.account_id = account_id
        self.partition = get_partition(region_name)
        self.clusters: Dict[str, Cluster] = OrderedDict()

    def create_cluster(
        self,
        deletion_protection_enabled: bool,
        tags: Optional[Dict[str, str]],
        client_token: Optional[str],
    ) -> Cluster:
        cluster = Cluster(
            self.region_name,
            self.account_id,
            deletion_protection_enabled,
            tags,
            client_token,
        )
        self.clusters[cluster.identifier] = cluster
        return cluster

    def get_cluster(
        self,
        identifier: str,
    ) -> Cluster:
        cluster = self.clusters.get(identifier)
        if cluster is None:
            raise ValidationException("invalid Cluster Id")

        return cluster


dsql_backends = BackendDict(
    AuroraDSQLBackend,
    "dsql",
    # currently botocore does not provide a dsql endpoint
    # https://github.com/boto/botocore/blob/e07cddc333fe4fb90efcd5d04324dd83f9cc3a57/botocore/data/endpoints.json
    use_boto3_regions=False,
    additional_regions=["us-east-1", "us-east-2"],
)



import logging

import localstack.services.stepfunctions.usage as UsageMetrics

from moto.stepfunctions.parser.asl.antlr.runtime.ASLParser import ASLParser
from moto.stepfunctions.parser.asl.component.common.query_language import (
    QueryLanguageMode,
)
from moto.stepfunctions.parser.asl.static_analyser.static_analyser import StaticAnalyser

LOG = logging.getLogger(__name__)


class UsageMetricsStaticAnalyser(StaticAnalyser):
    @staticmethod
    def process(definition: str) -> "UsageMetricsStaticAnalyser":
        analyser = UsageMetricsStaticAnalyser()
        try:
            analyser.analyse(definition=definition)

            if analyser.has_jsonata:
                UsageMetrics.jsonata_create_counter.increment()
            else:
                UsageMetrics.jsonpath_create_counter.increment()

            if analyser.has_variable_sampling:
                UsageMetrics.variables_create_counter.increment()
        except Exception as e:
            LOG.warning(
                "Failed to record Step Functions metrics from static analysis",
                exc_info=e,
            )
        return analyser

    def __init__(self):
        super().__init__()
        self.has_jsonata: bool = False
        self.has_variable_sampling = False

    def visitQuery_language_decl(self, ctx: ASLParser.Query_language_declContext):
        if self.has_jsonata:
            return

        query_language_mode_int = ctx.children[-1].getSymbol().type
        query_language_mode = QueryLanguageMode(value=query_language_mode_int)
        if query_language_mode == QueryLanguageMode.JSONata:
            self.has_jsonata = True

    def visitVariable_sample(self, ctx: ASLParser.Variable_sampleContext):
        self.has_variable_sampling = True

    def visitAssign_decl(self, ctx: ASLParser.Assign_declContext):
        self.has_variable_sampling = True



"""Exceptions raised by the securityhub service."""

from moto.core.exceptions import JsonRESTError


class SecurityHubClientError(JsonRESTError):
    code = 400


class _InvalidOperationException(SecurityHubClientError):
    def __init__(self, error_type: str, op: str, msg: str):
        super().__init__(
            error_type,
            "An error occurred (%s) when calling the %s operation: %s"
            % (error_type, op, msg),
        )


class InvalidInputException(_InvalidOperationException):
    def __init__(self, op: str, msg: str):
        super().__init__("InvalidInputException", op, msg)



"""Unit tests for workspacesweb-supported APIs."""

import re

import boto3

from moto import mock_aws

# See our Development Tips on writing tests for hints on how to write good tests:
# http://docs.getmoto.org/en/latest/docs/contributing/development_tips/tests.html

FAKE_SECURITY_GROUP_IDS = ["sg-0123456789abcdef0"]
FAKE_SUBNET_IDS = ["subnet-0123456789abcdef0", "subnet-abcdef0123456789"]
FAKE_TAGS = [
    {"Key": "TestKey", "Value": "TestValue"},
    {"Key": "TestKey2", "Value": "TestValue2"},
]
FAKE_VPC_ID = "vpc-0123456789abcdef0"
FAKE_KMS_KEY_ID = "abcd1234-5678-90ab-cdef-FAKEKEY"


@mock_aws
def test_create_network_settings():
    client = boto3.client("workspaces-web", region_name="eu-west-1")
    response = client.create_network_settings(
        securityGroupIds=FAKE_SECURITY_GROUP_IDS,
        subnetIds=FAKE_SUBNET_IDS,
        tags=FAKE_TAGS,
        vpcId=FAKE_VPC_ID,
    )
    network_settings_arn = response["networkSettingsArn"]
    arn_regex = r"^arn:aws:workspaces-web:eu-west-1:\d{12}:network-settings/[a-f0-9-]+$"
    assert re.match(arn_regex, network_settings_arn) is not None


@mock_aws
def test_list_network_settings():
    client = boto3.client("workspaces-web", region_name="ap-southeast-1")
    arn = client.create_network_settings(
        securityGroupIds=FAKE_SECURITY_GROUP_IDS,
        subnetIds=FAKE_SUBNET_IDS,
        tags=FAKE_TAGS,
        vpcId=FAKE_VPC_ID,
    )["networkSettingsArn"]
    resp = client.list_network_settings()
    assert resp["networkSettings"][0]["networkSettingsArn"] == arn
    assert len(resp["networkSettings"]) == 1


@mock_aws
def test_get_network_settings():
    client = boto3.client("workspaces-web", region_name="ap-southeast-1")
    arn = client.create_network_settings(
        securityGroupIds=FAKE_SECURITY_GROUP_IDS,
        subnetIds=FAKE_SUBNET_IDS,
        tags=FAKE_TAGS,
        vpcId=FAKE_VPC_ID,
    )["networkSettingsArn"]
    resp = client.get_network_settings(networkSettingsArn=arn)["networkSettings"]
    assert resp["networkSettingsArn"] == arn


@mock_aws
def test_delete_network_settings():
    client = boto3.client("workspaces-web", region_name="eu-west-1")
    arn = client.create_network_settings(
        securityGroupIds=FAKE_SECURITY_GROUP_IDS,
        subnetIds=FAKE_SUBNET_IDS,
        tags=FAKE_TAGS,
        vpcId=FAKE_VPC_ID,
    )["networkSettingsArn"]
    client.delete_network_settings(networkSettingsArn=arn)
    resp = client.list_network_settings()
    assert resp["networkSettings"] == [], "Expected no network settings"


@mock_aws
def test_create_browser_settings():
    client = boto3.client("workspaces-web", region_name="eu-west-1")
    response = client.create_browser_settings(
        additionalEncryptionContext={"Key1": "Value1", "Key2": "Value2"},
        browserPolicy="TestBrowserPolicy",
        clientToken="TestClient",
        customerManagedKey=FAKE_KMS_KEY_ID,
        tags=FAKE_TAGS,
    )
    browser_settings_arn = response["browserSettingsArn"]
    arn_regex = r"^arn:aws:workspaces-web:eu-west-1:\d{12}:browser-settings/[a-f0-9-]+$"
    assert re.match(arn_regex, browser_settings_arn) is not None


@mock_aws
def test_list_browser_settings():
    client = boto3.client("workspaces-web", region_name="eu-west-1")
    arn = client.create_browser_settings(
        additionalEncryptionContext={"Key1": "Value1", "Key2": "Value2"},
        browserPolicy="TestBrowserPolicy",
        clientToken="TestClient",
        customerManagedKey=FAKE_KMS_KEY_ID,
        tags=FAKE_TAGS,
    )["browserSettingsArn"]
    resp = client.list_browser_settings()
    assert resp["browserSettings"][0]["browserSettingsArn"] == arn


@mock_aws
def test_get_browser_settings():
    client = boto3.client("workspaces-web", region_name="eu-west-1")
    arn = client.create_browser_settings(
        additionalEncryptionContext={"Key1": "Value1", "Key2": "Value2"},
        browserPolicy="TestBrowserPolicy",
        clientToken="TestClient",
        customerManagedKey=FAKE_KMS_KEY_ID,
        tags=FAKE_TAGS,
    )["browserSettingsArn"]
    resp = client.get_browser_settings(browserSettingsArn=arn)["browserSettings"]
    assert resp["browserSettingsArn"] == arn


@mock_aws
def test_delete_browser_settings():
    client = boto3.client("workspaces-web", region_name="eu-west-1")
    arn = client.create_browser_settings(
        additionalEncryptionContext={"Key1": "Value1", "Key2": "Value2"},
        browserPolicy="TestBrowserPolicy",
        clientToken="TestClient",
        customerManagedKey=FAKE_KMS_KEY_ID,
        tags=FAKE_TAGS,
    )["browserSettingsArn"]
    client.delete_browser_settings(browserSettingsArn=arn)
    resp = client.list_browser_settings()
    assert resp["browserSettings"] == []


@mock_aws
def test_create_portal():
    client = boto3.client("workspaces-web", region_name="eu-west-1")
    resp = client.create_portal(
        additionalEncryptionContext={"Key1": "Value1", "Key2": "Value2"},
        authenticationType="Standard",
        clientToken="TestClient",
        customerManagedKey=FAKE_KMS_KEY_ID,
        displayName="TestDisplayName",
        instanceType="TestInstanceType",
        maxConcurrentSessions=5,
        tags=FAKE_TAGS,
    )
    assert resp["portalArn"]
    assert resp["portalEndpoint"]


@mock_aws
def test_list_portals():
    client = boto3.client("workspaces-web", region_name="eu-west-1")
    arn = client.create_portal(
        additionalEncryptionContext={"Key1": "Value1", "Key2": "Value2"},
        authenticationType="Standard",
        clientToken="TestClient",
        customerManagedKey=FAKE_KMS_KEY_ID,
        displayName="TestDisplayName",
        instanceType="TestInstanceType",
        maxConcurrentSessions=5,
        tags=FAKE_TAGS,
    )["portalArn"]
    resp = client.list_portals()
    assert resp["portals"][0]["portalArn"] == arn
    assert len(resp["portals"]) == 1
    assert resp["portals"][0]["authenticationType"] == "Standard"


@mock_aws
def test_get_portal():
    client = boto3.client("workspaces-web", region_name="eu-west-1")
    arn = client.create_portal(
        additionalEncryptionContext={"Key1": "Value1", "Key2": "Value2"},
        authenticationType="Standard",
        clientToken="TestClient",
        customerManagedKey=FAKE_KMS_KEY_ID,
        displayName="TestDisplayName",
        instanceType="TestInstanceType",
        maxConcurrentSessions=5,
        tags=FAKE_TAGS,
    )["portalArn"]
    resp = client.get_portal(portalArn=arn)["portal"]
    assert resp["portalArn"] == arn, f"Expected ARN {arn} in response"
    assert (
        resp["authenticationType"] == "Standard"
    ), "Expected authentication type in response"
    assert (
        resp["instanceType"] == "TestInstanceType"
    ), "Expected instance type in response"
    assert (
        resp["maxConcurrentSessions"] == 5
    ), "Expected max concurrent sessions in response"


@mock_aws
def test_delete_portal():
    client = boto3.client("workspaces-web", region_name="eu-west-1")
    arn = client.create_portal(
        additionalEncryptionContext={"Key1": "Value1", "Key2": "Value2"},
        authenticationType="Standard",
        clientToken="TestClient",
        customerManagedKey=FAKE_KMS_KEY_ID,
        displayName="TestDisplayName",
        instanceType="TestInstanceType",
        maxConcurrentSessions=5,
        tags=FAKE_TAGS,
    )["portalArn"]
    client.delete_portal(portalArn=arn)
    assert len(client.list_portals()["portals"]) == 0, "Expected no portals"


@mock_aws
def test_associate_browser_settings():
    client = boto3.client("workspaces-web", region_name="eu-west-1")
    browser_settings_arn = client.create_browser_settings(
        additionalEncryptionContext={"Key1": "Value1", "Key2": "Value2"},
        browserPolicy="TestBrowserPolicy",
        clientToken="TestClient",
        customerManagedKey=FAKE_KMS_KEY_ID,
        tags=FAKE_TAGS,
    )["browserSettingsArn"]
    portal_arn = client.create_portal(
        additionalEncryptionContext={"Key1": "Value1", "Key2": "Value2"},
        authenticationType="Standard",
        clientToken="TestClient",
        customerManagedKey=FAKE_KMS_KEY_ID,
        displayName="TestDisplayName",
        instanceType="TestInstanceType",
        maxConcurrentSessions=5,
        tags=FAKE_TAGS,
    )["portalArn"]
    client.associate_browser_settings(
        browserSettingsArn=browser_settings_arn, portalArn=portal_arn
    )
    resp = client.get_portal(portalArn=portal_arn)["portal"]
    assert resp["browserSettingsArn"] == browser_settings_arn
    resp = client.get_browser_settings(browserSettingsArn=browser_settings_arn)[
        "browserSettings"
    ]
    assert resp["associatedPortalArns"] == [portal_arn]


@mock_aws
def test_associate_network_settings():
    client = boto3.client("workspaces-web", region_name="eu-west-1")
    network_arn = client.create_network_settings(
        securityGroupIds=FAKE_SECURITY_GROUP_IDS,
        subnetIds=FAKE_SUBNET_IDS,
        tags=FAKE_TAGS,
        vpcId=FAKE_VPC_ID,
    )["networkSettingsArn"]
    portal_arn = client.create_portal(
        additionalEncryptionContext={"Key1": "Value1", "Key2": "Value2"},
        authenticationType="Standard",
        clientToken="TestClient",
        customerManagedKey=FAKE_KMS_KEY_ID,
        displayName="TestDisplayName",
        instanceType="TestInstanceType",
        maxConcurrentSessions=5,
        tags=FAKE_TAGS,
    )["portalArn"]
    client.associate_network_settings(
        networkSettingsArn=network_arn, portalArn=portal_arn
    )
    resp = client.get_portal(portalArn=portal_arn)["portal"]
    assert resp["networkSettingsArn"] == network_arn
    resp = client.get_network_settings(networkSettingsArn=network_arn)[
        "networkSettings"
    ]
    assert resp["associatedPortalArns"] == [portal_arn]


@mock_aws
def test_create_user_settings():
    client = boto3.client("workspaces-web", region_name="eu-west-1")
    resp = client.create_user_settings(
        copyAllowed="Disabled",
        pasteAllowed="Disabled",
        printAllowed="Disabled",
        uploadAllowed="Disabled",
        downloadAllowed="Disabled",
    )
    user_settings_arn = resp["userSettingsArn"]
    arn_regex = r"^arn:aws:workspaces-web:eu-west-1:\d{12}:user-settings/[a-f0-9-]+$"
    assert re.match(arn_regex, user_settings_arn) is not None


@mock_aws
def test_get_user_settings():
    client = boto3.client("workspaces-web", region_name="eu-west-1")
    resp = client.create_user_settings(
        copyAllowed="Enabled",
        pasteAllowed="Disabled",
        printAllowed="Enabled",
        uploadAllowed="Disabled",
        downloadAllowed="Enabled",
    )
    user_settings_arn = resp["userSettingsArn"]
    resp = client.get_user_settings(userSettingsArn=user_settings_arn)["userSettings"]
    assert resp["userSettingsArn"] == user_settings_arn
    assert resp["copyAllowed"] == "Enabled"
    assert resp["pasteAllowed"] == "Disabled"
    assert resp["printAllowed"] == "Enabled"
    assert resp["uploadAllowed"] == "Disabled"
    assert resp["downloadAllowed"] == "Enabled"


@mock_aws
def test_list_user_settings():
    client = boto3.client("workspaces-web", region_name="ap-southeast-1")
    arn = client.create_user_settings(
        copyAllowed="Disabled",
        pasteAllowed="Disabled",
        printAllowed="Disabled",
        uploadAllowed="Disabled",
        downloadAllowed="Disabled",
    )["userSettingsArn"]
    resp = client.list_user_settings()
    assert resp["userSettings"][0]["userSettingsArn"] == arn


@mock_aws
def test_delete_user_settings():
    client = boto3.client("workspaces-web", region_name="eu-west-1")
    arn = client.create_user_settings(
        copyAllowed="Disabled",
        pasteAllowed="Disabled",
        printAllowed="Disabled",
        uploadAllowed="Disabled",
        downloadAllowed="Disabled",
    )["userSettingsArn"]
    client.delete_user_settings(userSettingsArn=arn)
    assert len(client.list_user_settings()["userSettings"]) == 0


@mock_aws
def test_associate_user_settings():
    client = boto3.client("workspaces-web", region_name="eu-west-1")
    user_settings_arn = client.create_user_settings(
        copyAllowed="Disabled",
        pasteAllowed="Disabled",
        printAllowed="Disabled",
        uploadAllowed="Disabled",
        downloadAllowed="Disabled",
    )["userSettingsArn"]
    portal_arn = client.create_portal(
        additionalEncryptionContext={"Key1": "Value1", "Key2": "Value2"},
        authenticationType="Standard",
        clientToken="TestClient",
        customerManagedKey=FAKE_KMS_KEY_ID,
        displayName="TestDisplayName",
        instanceType="TestInstanceType",
        maxConcurrentSessions=5,
        tags=FAKE_TAGS,
    )["portalArn"]
    client.associate_user_settings(
        userSettingsArn=user_settings_arn, portalArn=portal_arn
    )
    resp = client.get_portal(portalArn=portal_arn)["portal"]
    assert resp["userSettingsArn"] == user_settings_arn
    resp = client.get_user_settings(userSettingsArn=user_settings_arn)["userSettings"]
    assert resp["associatedPortalArns"] == [portal_arn]


@mock_aws
def test_create_user_access_logging_settings():
    client = boto3.client("workspaces-web", region_name="ap-southeast-1")
    user_access_logging_settings_arn = client.create_user_access_logging_settings(
        kinesisStreamArn="arn:aws:kinesis:ap-southeast-1:123456789012:stream/TestStream",
    )["userAccessLoggingSettingsArn"]
    arn_regex = r"^arn:aws:workspaces-web:ap-southeast-1:\d{12}:user-access-logging-settings/[a-f0-9-]+$"
    assert re.match(arn_regex, user_access_logging_settings_arn) is not None


@mock_aws
def test_get_user_access_logging_settings():
    client = boto3.client("workspaces-web", region_name="eu-west-1")
    user_access_logging_settings_arn = client.create_user_access_logging_settings(
        kinesisStreamArn="arn:aws:kinesis:ap-southeast-1:123456789012:stream/TestStream",
    )["userAccessLoggingSettingsArn"]
    resp = client.get_user_access_logging_settings(
        userAccessLoggingSettingsArn=user_access_logging_settings_arn
    )["userAccessLoggingSettings"]
    assert resp["userAccessLoggingSettingsArn"] == user_access_logging_settings_arn
    assert (
        resp["kinesisStreamArn"]
        == "arn:aws:kinesis:ap-southeast-1:123456789012:stream/TestStream"
    )


@mock_aws
def test_list_user_access_logging_settings():
    client = boto3.client("workspaces-web", region_name="eu-west-1")
    arn = client.create_user_access_logging_settings(
        kinesisStreamArn="arn:aws:kinesis:ap-southeast-1:123456789012:stream/TestStream",
    )["userAccessLoggingSettingsArn"]
    resp = client.list_user_access_logging_settings()
    assert resp["userAccessLoggingSettings"][0]["userAccessLoggingSettingsArn"] == arn


@mock_aws
def test_delete_user_access_logging_settings():
    client = boto3.client("workspaces-web", region_name="ap-southeast-1")
    arn = client.create_user_access_logging_settings(
        kinesisStreamArn="arn:aws:kinesis:ap-southeast-1:123456789012:stream/TestStream",
    )["userAccessLoggingSettingsArn"]
    client.delete_user_access_logging_settings(userAccessLoggingSettingsArn=arn)
    resp = client.list_user_access_logging_settings()
    assert resp["userAccessLoggingSettings"] == []


@mock_aws
def test_associate_user_access_logging_settings():
    client = boto3.client("workspaces-web", region_name="eu-west-1")
    user_access_logging_settings_arn = client.create_user_access_logging_settings(
        kinesisStreamArn="arn:aws:kinesis:ap-southeast-1:123456789012:stream/TestStream",
    )["userAccessLoggingSettingsArn"]
    portal_arn = client.create_portal(
        additionalEncryptionContext={"Key1": "Value1", "Key2": "Value2"},
        authenticationType="Standard",
        clientToken="TestClient",
        customerManagedKey=FAKE_KMS_KEY_ID,
        displayName="TestDisplayName",
        instanceType="TestInstanceType",
        maxConcurrentSessions=5,
        tags=FAKE_TAGS,
    )["portalArn"]
    client.associate_user_access_logging_settings(
        userAccessLoggingSettingsArn=user_access_logging_settings_arn,
        portalArn=portal_arn,
    )
    resp = client.get_portal(portalArn=portal_arn)["portal"]
    assert resp["userAccessLoggingSettingsArn"] == user_access_logging_settings_arn
    resp = client.get_user_access_logging_settings(
        userAccessLoggingSettingsArn=user_access_logging_settings_arn
    )["userAccessLoggingSettings"]
    assert resp["associatedPortalArns"] == [portal_arn]


@mock_aws
def test_list_tags_for_resource():
    client = boto3.client("workspaces-web", region_name="ap-southeast-1")

    OTHER_FAKE_TAGS = [
        {"Key": "FAKEKEY1", "Value": "FAKEVALUE1"},
        {"Key": "FAKEKEY2", "Value": "FAKEVALUE2"},
    ]

    arns = []

    arns.append(
        client.create_portal(
            additionalEncryptionContext={"Key1": "Encryption", "Key2": "Context"},
            authenticationType="Standard",
            clientToken="TestClient",
            customerManagedKey=FAKE_KMS_KEY_ID,
            displayName="TestDisplayName",
            instanceType="TestInstanceType",
            maxConcurrentSessions=5,
            tags=FAKE_TAGS,
        )["portalArn"]
    )

    arns.append(
        client.create_network_settings(
            securityGroupIds=FAKE_SECURITY_GROUP_IDS,
            subnetIds=FAKE_SUBNET_IDS,
            tags=FAKE_TAGS,
            vpcId=FAKE_VPC_ID,
        )["networkSettingsArn"]
    )

    arns.append(
        client.create_user_settings(
            copyAllowed="Disabled",
            pasteAllowed="Disabled",
            printAllowed="Disabled",
            uploadAllowed="Disabled",
            downloadAllowed="Disabled",
            tags=FAKE_TAGS,
        )["userSettingsArn"]
    )

    arns.append(
        client.create_user_access_logging_settings(
            kinesisStreamArn="arn:aws:kinesis:ap-southeast-1:123456789012:stream/TestStream",
            tags=FAKE_TAGS,
        )["userAccessLoggingSettingsArn"]
    )

    arns.append(
        client.create_browser_settings(
            additionalEncryptionContext={"Key1": "Value1", "Key2": "Value2"},
            browserPolicy="TestBrowserPolicy",
            clientToken="TestClient",
            customerManagedKey=FAKE_KMS_KEY_ID,
            tags=FAKE_TAGS,
        )["browserSettingsArn"]
    )

    for arn in arns:
        resp = client.list_tags_for_resource(resourceArn=arn)
        assert resp["tags"] == FAKE_TAGS

        client.tag_resource(resourceArn=arn, tags=OTHER_FAKE_TAGS)
        resp = client.list_tags_for_resource(resourceArn=arn)
        assert resp["tags"] == FAKE_TAGS + OTHER_FAKE_TAGS

        client.untag_resource(resourceArn=arn, tagKeys=["FAKEKEY1", "TestKey"])
        resp = client.list_tags_for_resource(resourceArn=arn)
        assert resp["tags"] == [
            {"Key": "TestKey2", "Value": "TestValue2"},
            {"Key": "FAKEKEY2", "Value": "FAKEVALUE2"},
        ]



from uuid import uuid4

import boto3
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa

from moto import mock_aws
from tests import aws_verified

from .cloudfront_test_scaffolding import minimal_dist_custom_config

private_key = rsa.generate_private_key(public_exponent=65537, key_size=2048)
public_key = private_key.public_key()
pem = public_key.public_bytes(
    encoding=serialization.Encoding.PEM,
    format=serialization.PublicFormat.SubjectPublicKeyInfo,
)


@aws_verified
def test_create_public_key():
    client = boto3.client("cloudfront", region_name="us-east-1")

    key_id = etag = None
    try:
        public_key_config = {
            "CallerReference": (str(uuid4())),
            "Name": (str(uuid4())),
            "EncodedKey": pem.decode("utf-8"),
        }
        resp = client.create_public_key(PublicKeyConfig=public_key_config)
        # uppercase+digits{14}
        key_id = resp["PublicKey"]["Id"]
        etag = resp["ETag"]
        public_key_config["Comment"] = ""

        assert resp["ResponseMetadata"]["HTTPStatusCode"] == 201

        assert (
            resp["Location"]
            == f"https://cloudfront.amazonaws.com/2020-05-31/public-key/{key_id}"
        )
        assert resp["PublicKey"]["CreatedTime"]
        assert resp["PublicKey"]["PublicKeyConfig"] == public_key_config

        get_key = client.get_public_key(Id=key_id)

        assert get_key["ETag"] == etag
        assert get_key["PublicKey"]["Id"] == key_id
        assert get_key["PublicKey"]["PublicKeyConfig"] == public_key_config
    finally:
        if key_id:
            client.delete_public_key(Id=key_id, IfMatch=etag)


@mock_aws
def test_list_public_keys():
    client = boto3.client("cloudfront", region_name="us-east-1")

    key_list = client.list_public_keys()["PublicKeyList"]
    assert key_list == {"MaxItems": 100, "Quantity": 0}

    key = client.create_public_key(
        PublicKeyConfig={
            "CallerReference": "someref",
            "Name": "somekey",
            "EncodedKey": "some data",
        }
    )["PublicKey"]
    key_created = key["CreatedTime"]
    key_id = key["Id"]

    key_list = client.list_public_keys()["PublicKeyList"]
    assert key_list["Quantity"] == 1
    assert key_list["Items"] == [
        {
            "Id": key_id,
            "Name": "somekey",
            "CreatedTime": key_created,
            "EncodedKey": "some data\n",
            "Comment": "",
        }
    ]


@mock_aws
def test_create_public_key_group():
    client = boto3.client("cloudfront", region_name="us-east-1")

    key_id1 = client.create_public_key(
        PublicKeyConfig={
            "CallerReference": str(uuid4()),
            "Name": str(uuid4()),
            "EncodedKey": pem.decode("utf-8"),
        }
    )["PublicKey"]["Id"]

    resp = client.create_key_group(
        KeyGroupConfig={"Name": "mygroup", "Items": [key_id1]}
    )
    key_group_id = resp["KeyGroup"]["Id"]  # uuid
    assert (
        resp["Location"]
        == f"https://cloudfront.amazonaws.com/2020-05-31/key-group/{key_group_id}"
    )
    assert resp["KeyGroup"]["KeyGroupConfig"] == {"Name": "mygroup", "Items": [key_id1]}

    get_group = client.get_key_group(Id=key_group_id)
    assert get_group["ETag"] == resp["ETag"]
    assert get_group["KeyGroup"]["Id"] == key_group_id
    assert get_group["KeyGroup"]["KeyGroupConfig"]["Name"] == "mygroup"
    assert get_group["KeyGroup"]["KeyGroupConfig"]["Items"] == [key_id1]

    groups = client.list_key_groups()["KeyGroupList"]
    assert groups["Quantity"] == 1
    assert groups["Items"][0]["KeyGroup"]["Id"] == key_group_id
    assert groups["Items"][0]["KeyGroup"]["KeyGroupConfig"]["Name"] == "mygroup"

    # We can't delete the key at this point
    # Either deregister the key, or delete the group
    # Deleting the group automatically deregisters the key (but does not delete it)


@mock_aws
def test_create_distribution_with_key_groups():
    client = boto3.client("cloudfront", region_name="us-east-1")

    key_id1 = client.create_public_key(
        PublicKeyConfig={
            "CallerReference": str(uuid4()),
            "Name": str(uuid4()),
            "EncodedKey": pem.decode("utf-8"),
        }
    )["PublicKey"]["Id"]

    resp = client.create_key_group(
        KeyGroupConfig={"Name": "mygroup", "Items": [key_id1]}
    )
    key_group_id = resp["KeyGroup"]["Id"]  # uuid

    config = minimal_dist_custom_config("sth")
    config["DefaultCacheBehavior"]["TrustedKeyGroups"] = {
        "Enabled": True,
        "Quantity": 1,
        "Items": [key_group_id],
    }
    dist = client.create_distribution(DistributionConfig=config)["Distribution"]
    config = dist["DistributionConfig"]
    key_groups = config["DefaultCacheBehavior"]["TrustedKeyGroups"]
    assert key_groups == {"Enabled": True, "Quantity": 1, "Items": [key_group_id]}



"""Handles incoming s3tables requests, invokes methods, returns responses."""

import json
from typing import Any, Dict
from urllib.parse import unquote

from moto.core.common_types import TYPE_RESPONSE
from moto.core.responses import BaseResponse

from .models import S3TablesBackend, s3tables_backends


class S3TablesResponse(BaseResponse):
    """Handler for S3Tables requests and responses."""

    def __init__(self) -> None:
        super().__init__(service_name="s3tables")
        self.default_response_headers = {"Content-Type": "application/json"}

    @property
    def s3tables_backend(self) -> S3TablesBackend:
        """Return backend instance specific for this region."""
        return s3tables_backends[self.current_account][self.region]

    def create_table_bucket(self) -> TYPE_RESPONSE:
        name = json.loads(self.body)["name"]
        bucket = self.s3tables_backend.create_table_bucket(
            name=name,
        )
        return 200, self.default_response_headers, json.dumps(dict(arn=bucket.arn))

    def list_table_buckets(self) -> TYPE_RESPONSE:
        params = self._get_params()
        prefix = params.get("prefix")
        continuation_token = params.get("continuationToken")
        max_buckets = params.get("maxBuckets")
        table_buckets, continuation_token = self.s3tables_backend.list_table_buckets(
            prefix=prefix,
            continuation_token=continuation_token,
            max_buckets=int(max_buckets) if max_buckets else None,
        )

        body: Dict[str, Any] = {
            "tableBuckets": [
                {
                    "arn": b.arn,
                    "name": b.name,
                    "ownerAccountId": b.account_id,
                    "createdAt": b.creation_date.isoformat(),
                }
                for b in table_buckets
            ]
        }
        if continuation_token:
            body.update(continuationToken=continuation_token)

        return 200, self.default_response_headers, json.dumps(body)

    def get_table_bucket(self) -> TYPE_RESPONSE:
        _, table_bucket_arn = self.raw_path.lstrip("/").split("/")
        table_bucket_arn = unquote(table_bucket_arn)
        bucket = self.s3tables_backend.get_table_bucket(
            table_bucket_arn=table_bucket_arn,
        )

        return (
            200,
            self.default_response_headers,
            json.dumps(
                dict(
                    arn=bucket.arn,
                    name=bucket.name,
                    ownerAccountId=bucket.account_id,
                    createdAt=bucket.creation_date.isoformat(),
                )
            ),
        )

    def delete_table_bucket(self) -> TYPE_RESPONSE:
        _, table_bucket_arn = self.raw_path.lstrip("/").split("/")
        table_bucket_arn = unquote(table_bucket_arn)
        self.s3tables_backend.delete_table_bucket(
            table_bucket_arn=table_bucket_arn,
        )

        return 204, {}, ""

    def create_namespace(self) -> TYPE_RESPONSE:
        _, table_bucket_arn = self.raw_path.lstrip("/").split("/")
        table_bucket_arn = unquote(table_bucket_arn)
        name = json.loads(self.body)["namespace"][0]
        namespace = self.s3tables_backend.create_namespace(
            table_bucket_arn=table_bucket_arn,
            namespace=name,
        )
        return (
            200,
            self.default_response_headers,
            json.dumps(
                dict(tableBucketArn=table_bucket_arn, namespace=[namespace.name])
            ),
        )

    def list_namespaces(self) -> TYPE_RESPONSE:
        _, table_bucket_arn = self.raw_path.lstrip("/").split("/")
        table_bucket_arn = unquote(table_bucket_arn)

        params = self._get_params()
        continuation_token = params.get("continuationToken")
        max_namespaces = params.get("maxNamespaces")
        prefix = params.get("prefix")

        namespaces, continuation_token = self.s3tables_backend.list_namespaces(
            table_bucket_arn=table_bucket_arn,
            prefix=prefix,
            continuation_token=continuation_token,
            max_namespaces=int(max_namespaces) if max_namespaces else None,
        )

        body: Dict[str, Any] = {
            "namespaces": [
                {
                    "namespace": [ns.name],
                    "createdAt": ns.creation_date.isoformat(),
                    "createdBy": ns.created_by,
                    "ownerAccountId": ns.account_id,
                }
                for ns in namespaces
            ]
        }
        if continuation_token:
            body.update(continuationToken=continuation_token)

        return 200, self.default_response_headers, json.dumps(body)

    def get_namespace(self) -> TYPE_RESPONSE:
        _, table_bucket_arn, name = self.raw_path.lstrip("/").split("/")
        table_bucket_arn = unquote(table_bucket_arn)
        namespace = self.s3tables_backend.get_namespace(
            table_bucket_arn=table_bucket_arn,
            namespace=name,
        )
        return (
            200,
            self.default_response_headers,
            json.dumps(
                dict(
                    namespace=[namespace.name],
                    createdAt=namespace.creation_date.isoformat(),
                    createdBy=namespace.created_by,
                    ownerAccountId=namespace.account_id,
                )
            ),
        )

    def delete_namespace(self) -> TYPE_RESPONSE:
        _, table_bucket_arn, namespace = self.raw_path.lstrip("/").split("/")
        table_bucket_arn = unquote(table_bucket_arn)
        self.s3tables_backend.delete_namespace(
            table_bucket_arn=table_bucket_arn,
            namespace=namespace,
        )
        return 204, self.default_response_headers, ""

    def create_table(self) -> TYPE_RESPONSE:
        _, table_bucket_arn, namespace = self.raw_path.lstrip("/").split("/")
        table_bucket_arn = unquote(table_bucket_arn)
        body = json.loads(self.body)
        name = body["name"]
        format = body["format"]
        table = self.s3tables_backend.create_table(
            table_bucket_arn=table_bucket_arn,
            namespace=namespace,
            name=name,
            format=format,
        )
        return (
            200,
            self.default_response_headers,
            json.dumps(dict(tableARN=table.arn, versionToken=table.version_token)),
        )

    def get_table(self) -> TYPE_RESPONSE:
        _, table_bucket_arn, namespace, name = self.raw_path.lstrip("/").split("/")
        table_bucket_arn = unquote(table_bucket_arn)
        table = self.s3tables_backend.get_table(
            table_bucket_arn=table_bucket_arn,
            namespace=namespace,
            name=name,
        )
        return (
            200,
            self.default_response_headers,
            json.dumps(
                dict(
                    name=table.name,
                    type=table.type,
                    tableARN=table.arn,
                    namespace=[namespace],
                    versionToken=table.version_token,
                    metadataLocation=table.metadata_location,
                    warehouseLocation=table.warehouse_location,
                    createdAt=table.creation_date.isoformat(),
                    createdBy=table.account_id,
                    managedByService=table.managed_by_service,
                    modifiedAt=table.last_modified.isoformat(),
                    modifiedBy=table.modified_by,
                    ownerAccountId=table.account_id,
                    format=table.format,
                )
            ),
        )

    def list_tables(self) -> TYPE_RESPONSE:
        _, table_bucket_arn = self.raw_path.lstrip("/").split("/")
        table_bucket_arn = unquote(table_bucket_arn)
        params = self._get_params()
        namespace = params.get("namespace")
        prefix = params.get("prefix")
        continuation_token = params.get("continuationToken")
        max_tables = params.get("maxTables")
        tables, continuation_token = self.s3tables_backend.list_tables(
            table_bucket_arn=table_bucket_arn,
            namespace=namespace,
            prefix=prefix,
            continuation_token=continuation_token,
            max_tables=int(max_tables) if max_tables else None,
        )
        body: Dict[str, Any] = {
            "tables": [
                {
                    "namespace": [table.namespace],
                    "name": table.name,
                    "createdAt": table.creation_date.isoformat(),
                    "modifiedAt": table.last_modified.isoformat(),
                }
                for table in tables
            ]
        }

        if continuation_token:
            body.update(continuationToken=continuation_token)

        return 200, self.default_response_headers, json.dumps(body)

    def delete_table(self) -> TYPE_RESPONSE:
        _, table_bucket_arn, namespace, name = self.raw_path.lstrip("/").split("/")
        table_bucket_arn = unquote(table_bucket_arn)
        params = self._get_params()
        version_token = params.get("versionToken")
        self.s3tables_backend.delete_table(
            table_bucket_arn=table_bucket_arn,
            namespace=namespace,
            name=name,
            version_token=version_token,
        )
        return 204, {}, ""

    def get_table_metadata_location(self) -> TYPE_RESPONSE:
        _, table_bucket_arn, namespace, name, _ = self.raw_path.lstrip("/").split("/")
        table_bucket_arn = unquote(table_bucket_arn)
        table = self.s3tables_backend.get_table(
            table_bucket_arn=table_bucket_arn,
            namespace=namespace,
            name=name,
        )
        return (
            200,
            self.default_response_headers,
            json.dumps(
                dict(
                    versionToken=table.version_token,
                    metadataLocation=table.metadata_location,
                    warehouseLocation=table.warehouse_location,
                )
            ),
        )

    def update_table_metadata_location(self) -> TYPE_RESPONSE:
        _, table_bucket_arn, namespace, name, _ = self.raw_path.lstrip("/").split("/")
        table_bucket_arn = unquote(table_bucket_arn)
        body = json.loads(self.body)
        metadata_location = body["metadataLocation"]
        version_token = body["versionToken"]
        table = self.s3tables_backend.update_table_metadata_location(
            table_bucket_arn=table_bucket_arn,
            namespace=namespace,
            name=name,
            version_token=version_token,
            metadata_location=metadata_location,
        )
        return (
            200,
            self.default_response_headers,
            json.dumps(
                dict(
                    name=table.name,
                    tableArn=table.arn,
                    namespace=namespace,
                    versionToken=table.version_token,
                    metadataLocation=table.metadata_location,
                )
            ),
        )

    def rename_table(self) -> TYPE_RESPONSE:
        _, table_bucket_arn, namespace, name, _ = self.raw_path.lstrip("/").split("/")
        table_bucket_arn = unquote(table_bucket_arn)
        body = json.loads(self.body)
        version_token = body.get("versionToken")
        new_namespace_name = body.get("newNamespaceName")
        new_name = body.get("newName")
        self.s3tables_backend.rename_table(
            table_bucket_arn=table_bucket_arn,
            namespace=namespace,
            name=name,
            new_namespace_name=new_namespace_name,
            new_name=new_name,
            version_token=version_token,
        )
        return 200, {}, ""



import datetime
import re

from moto.stepfunctions.parser.api import ExecutionFailedEventDetails, HistoryEventType
from moto.stepfunctions.parser.asl.component.common.error_name.failure_event import (
    FailureEvent,
    FailureEventException,
)
from moto.stepfunctions.parser.asl.component.common.error_name.states_error_name import (
    StatesErrorName,
)
from moto.stepfunctions.parser.asl.component.common.error_name.states_error_name_type import (
    StatesErrorNameType,
)
from moto.stepfunctions.parser.asl.component.common.variable_sample import (
    VariableSample,
)
from moto.stepfunctions.parser.asl.component.state.wait.wait_function.timestamp import (
    TIMESTAMP_FORMAT,
    TIMESTAMP_PATTERN,
)
from moto.stepfunctions.parser.asl.component.state.wait.wait_function.wait_function import (
    WaitFunction,
)
from moto.stepfunctions.parser.asl.eval.environment import Environment
from moto.stepfunctions.parser.asl.eval.event.event_detail import EventDetails
from moto.stepfunctions.parser.asl.utils.json_path import extract_json


class TimestampPath(WaitFunction):
    # TimestampPath
    # An absolute time to state_wait until beginning the state specified in the Next field,
    # specified using a path from the state's input data.

    def __init__(self, path: str):
        self.path: str = path

    def _create_failure_event(
        self, env: Environment, timestamp_str: str
    ) -> FailureEvent:
        return FailureEvent(
            env=env,
            error_name=StatesErrorName(typ=StatesErrorNameType.StatesRuntime),
            event_type=HistoryEventType.ExecutionFailed,
            event_details=EventDetails(
                executionFailedEventDetails=ExecutionFailedEventDetails(
                    error=StatesErrorNameType.StatesRuntime.to_name(),
                    cause=f"The TimestampPath parameter does not reference a valid ISO-8601 extended offset date-time format string: {self.path} == {timestamp_str}",
                )
            ),
        )

    def _compute_delta_seconds(self, env: Environment, timestamp_str: str):
        try:
            if not re.match(TIMESTAMP_PATTERN, timestamp_str):
                raise FailureEventException(
                    self._create_failure_event(env, timestamp_str)
                )

            # anything lower than seconds is truncated
            processed_timestamp = timestamp_str.rsplit(".", 2)[0]
            # add back the "Z" suffix if we removed it
            if not processed_timestamp.endswith("Z"):
                processed_timestamp = f"{processed_timestamp}Z"
            timestamp = datetime.datetime.strptime(
                processed_timestamp, TIMESTAMP_FORMAT
            )
        except Exception:
            raise FailureEventException(self._create_failure_event(env, timestamp_str))

        delta = timestamp - datetime.datetime.now()
        delta_sec = int(delta.total_seconds())
        return delta_sec

    def _get_wait_seconds(self, env: Environment) -> int:
        inp = env.stack[-1]
        timestamp_str: str = extract_json(self.path, inp)
        delta_sec = self._compute_delta_seconds(env=env, timestamp_str=timestamp_str)
        return delta_sec


class TimestampPathVar(TimestampPath):
    variable_sample: VariableSample

    def __init__(self, variable_sample: VariableSample):
        super().__init__(path=variable_sample.expression)
        self.variable_sample = variable_sample

    def _get_wait_seconds(self, env: Environment) -> int:
        self.variable_sample.eval(env=env)
        timestamp_str = env.stack.pop()
        delta_sec = self._compute_delta_seconds(env=env, timestamp_str=timestamp_str)
        return delta_sec



import json
from uuid import uuid4

import boto3
import pytest

from tests import aws_verified


@aws_verified
@pytest.mark.aws_verified
def test_tagging():
    logs_client = boto3.client("logs", region_name="us-east-1")
    cf_client = boto3.client("cloudformation", region_name="us-east-1")

    log_group_name = f"/moto/test/{str(uuid4())}"

    template = {
        "AWSTemplateFormatVersion": "2010-09-09",
        "Resources": {
            "testGroup": {
                "Type": "AWS::Logs::LogGroup",
                "Properties": {
                    "LogGroupName": log_group_name,
                    "Tags": [{"Key": "foo", "Value": "bar"}],
                },
            }
        },
    }
    template_json = json.dumps(template)
    stack_name = f"moto-test-{str(uuid4())[0:6]}"
    cf_client.create_stack(
        StackName=stack_name,
        TemplateBody=template_json,
    )
    waiter = cf_client.get_waiter("stack_create_complete")
    waiter.wait(StackName=stack_name)

    group = logs_client.describe_log_groups(logGroupNamePrefix=log_group_name)[
        "logGroups"
    ][0]

    tags = logs_client.list_tags_for_resource(resourceArn=group["logGroupArn"])["tags"]
    assert tags["foo"] == "bar"

    cf_client.delete_stack(StackName=stack_name)
    waiter = cf_client.get_waiter("stack_delete_complete")
    waiter.wait(StackName=stack_name)
    assert (
        logs_client.describe_log_groups(logGroupNamePrefix=log_group_name)["logGroups"]
        == []
    )



import re
from typing import TYPE_CHECKING, Iterator, List, Union

if TYPE_CHECKING:
    from ..models import FakeSecret


def name_filter(secret: "FakeSecret", names: List[str]) -> bool:
    return _matcher(names, [secret.name])


def description_filter(secret: "FakeSecret", descriptions: List[str]) -> bool:
    if not secret.description:
        return False
    # The documentation states that this search uses `Prefix match`
    # But actual testing determines that it uses the same approach to the `all_filter`:
    # 'Breaks the filter value string into words and then searches all attributes for matches.'
    return _matcher(
        descriptions, [secret.description], match_prefix=False, case_sensitive=False
    )


def tag_key(secret: "FakeSecret", tag_keys: List[str]) -> bool:
    if not secret.tags:
        return False
    return _matcher(tag_keys, [tag["Key"] for tag in secret.tags])


def tag_value(secret: "FakeSecret", tag_values: List[str]) -> bool:
    if not secret.tags:
        return False
    return _matcher(tag_values, [tag["Value"] for tag in secret.tags])


def filter_all(secret: "FakeSecret", values: List[str]) -> bool:
    attributes = [secret.name]
    if secret.description:
        attributes.append(secret.description)
    if secret.tags:
        attributes += [tag["Key"] for tag in secret.tags] + [
            tag["Value"] for tag in secret.tags
        ]

    return _matcher(values, attributes, match_prefix=False, case_sensitive=False)


def _matcher(
    patterns: List[str],
    strings: List[str],
    match_prefix: bool = True,
    case_sensitive: bool = True,
) -> bool:
    for pattern in [p for p in patterns if p.startswith("!")]:
        for string in strings:
            if not _match_pattern(
                pattern[1:], string, match_prefix, case_sensitive=case_sensitive
            ):
                return True

    for pattern in [p for p in patterns if not p.startswith("!")]:
        for string in strings:
            if _match_pattern(
                pattern, string, match_prefix, case_sensitive=case_sensitive
            ):
                return True
    return False


def _match_pattern(
    pattern: str, value: str, match_prefix: bool = True, case_sensitive: bool = True
) -> bool:
    if match_prefix:
        if not case_sensitive:
            return value.lower().startswith(pattern.lower())
        else:
            return value.startswith(pattern)
    else:
        pattern_words = split_words(pattern)
        if not pattern_words:
            return False
        value_words = split_words(value)
        if not case_sensitive:
            pattern_words = [p.lower() for p in pattern_words]
            value_words = [v.lower() for v in value_words]
        for pattern_word in pattern_words:
            # all words in value must start with pattern_word
            if not any(
                value_word.startswith(pattern_word) for value_word in value_words
            ):
                return False
    return True


def split_words(s: str) -> List[str]:
    """
    Secrets are split by special characters first (/, +, _, etc)
    Partial results are then split again by UpperCasing
    """
    special_chars = ["/", "-", "_", "+", "=", ".", "@"]

    if s in special_chars:
        # Special case: this does not return any values
        return []

    for char in special_chars:
        if char in s:
            others = special_chars.copy()
            others.remove(char)
            contains_other = any([c in s for c in others])
            if contains_other:
                # Secret contains two different characters, i.e. my/secret+value
                # Values like this will not be split
                return [s]
            else:
                return list(split_by_uppercase(s.split(char)))
    return list(split_by_uppercase(s))


def split_by_uppercase(s: Union[str, List[str]]) -> Iterator[str]:
    """
    Split a string into words. Words are recognized by upper case letters, i.e.:
    test   -> [test]
    MyTest -> [My, Test]
    """
    if isinstance(s, str):
        for x in re.split(r"([^a-z][a-z]+)", s):
            if x:
                yield x.strip()
    else:
        for word in s:
            yield from split_by_uppercase(word)



from moto.utilities.id_generator import (
    TAG_KEY_CUSTOM_ID,
    ExistingIds,
    ResourceIdentifier,
    Tags,
    moto_id,
    moto_id_manager,
)

ACCOUNT = "account"
REGION = "us-east-1"
RESOURCE_NAME = "my-resource"

CUSTOM_ID = "custom"
GENERIC_ID = "generated"
TAG_ID = "fromTag"
SERVICE = "test-service"
RESOURCE = "test-resource"


@moto_id
def generate_test_id(
    resource_identifier: ResourceIdentifier,
    existing_ids: ExistingIds = None,
    tags: Tags = None,
):
    return GENERIC_ID


class TestResourceIdentifier(ResourceIdentifier):
    __test__ = False  # Prevent pytest discovery

    service = SERVICE
    resource = RESOURCE

    def generate(self, existing_ids: ExistingIds = None, tags: Tags = None) -> str:
        return generate_test_id(
            resource_identifier=self, existing_ids=existing_ids, tags=tags
        )


def test_generate_with_no_resource_identifier():
    generated_id = generate_test_id(None)
    assert generated_id == GENERIC_ID


def test_generate_with_matching_resource_identifier(set_custom_id):
    resource_identifier = TestResourceIdentifier(ACCOUNT, REGION, RESOURCE_NAME)

    set_custom_id(resource_identifier, CUSTOM_ID)

    generated_id = generate_test_id(resource_identifier=resource_identifier)
    assert generated_id == CUSTOM_ID


def test_generate_with_non_matching_resource_identifier(set_custom_id):
    resource_identifier = TestResourceIdentifier(ACCOUNT, REGION, RESOURCE_NAME)
    resource_identifier_2 = TestResourceIdentifier(ACCOUNT, REGION, "non-matching")

    set_custom_id(resource_identifier, CUSTOM_ID)

    generated_id = generate_test_id(resource_identifier=resource_identifier_2)
    assert generated_id == GENERIC_ID


def test_generate_with_custom_id_tag():
    resource_identifier = TestResourceIdentifier(ACCOUNT, REGION, RESOURCE_NAME)

    generated_id = generate_test_id(
        resource_identifier=resource_identifier, tags={TAG_KEY_CUSTOM_ID: TAG_ID}
    )
    assert generated_id == TAG_ID


def test_generate_with_custom_id_tag_has_priority(set_custom_id):
    resource_identifier = TestResourceIdentifier(ACCOUNT, REGION, RESOURCE_NAME)

    set_custom_id(resource_identifier, CUSTOM_ID)
    generated_id = generate_test_id(
        resource_identifier=resource_identifier, tags={TAG_KEY_CUSTOM_ID: TAG_ID}
    )
    assert generated_id == TAG_ID


def test_generate_with_existing_id(set_custom_id):
    resource_identifier = TestResourceIdentifier(ACCOUNT, REGION, RESOURCE_NAME)

    set_custom_id(resource_identifier, CUSTOM_ID)
    generated_id = generate_test_id(
        resource_identifier=resource_identifier, existing_ids=[CUSTOM_ID]
    )
    assert generated_id == GENERIC_ID


def test_generate_with_tags_and_existing_id(set_custom_id):
    resource_identifier = TestResourceIdentifier(ACCOUNT, REGION, RESOURCE_NAME)

    generated_id = generate_test_id(
        resource_identifier=resource_identifier,
        existing_ids=[TAG_ID],
        tags={TAG_KEY_CUSTOM_ID: TAG_ID},
    )
    assert generated_id == GENERIC_ID


def test_generate_with_tags_fallback(set_custom_id):
    resource_identifier = TestResourceIdentifier(ACCOUNT, REGION, RESOURCE_NAME)

    set_custom_id(resource_identifier, CUSTOM_ID)
    generated_id = generate_test_id(
        resource_identifier=resource_identifier,
        existing_ids=[TAG_ID],
        tags={TAG_KEY_CUSTOM_ID: TAG_ID},
    )
    assert generated_id == CUSTOM_ID


def test_set_custom_id_lifecycle():
    resource_identifier = TestResourceIdentifier(ACCOUNT, REGION, RESOURCE_NAME)

    moto_id_manager.set_custom_id(resource_identifier, CUSTOM_ID)

    found_id = moto_id_manager.get_custom_id(resource_identifier)
    assert found_id == CUSTOM_ID

    moto_id_manager.unset_custom_id(resource_identifier)

    found_id = moto_id_manager.get_custom_id(resource_identifier)
    assert found_id is None


def test_set_custom_id_name_is_not_set():
    resource_identifier = TestResourceIdentifier(ACCOUNT, REGION, None)
    moto_id_manager.set_custom_id(resource_identifier, CUSTOM_ID)

    assert moto_id_manager._custom_ids == {}



from __future__ import annotations

import abc
from typing import Optional

from moto.stepfunctions.parser.asl.component.component import Component


class ErrorName(Component, abc.ABC):
    error_name: Optional[str]

    def __init__(self, error_name: Optional[str]):
        self.error_name = error_name

    def matches(self, error_name: Optional[str]) -> bool:
        return self.error_name == error_name

    def __eq__(self, other):
        if isinstance(other, ErrorName):
            return self.matches(other.error_name)
        return False



import abc
from collections import OrderedDict
from threading import Event, Lock
from typing import Dict, Optional

from moto.stepfunctions.parser.api import ActivityDoesNotExist, Arn
from moto.stepfunctions.parser.backend.activity import Activity, ActivityTask
from moto.stepfunctions.parser.utils import long_uid

CallbackId = str


class CallbackOutcome(abc.ABC):
    callback_id: CallbackId

    def __init__(self, callback_id: str):
        self.callback_id = callback_id


class CallbackOutcomeSuccess(CallbackOutcome):
    output: str

    def __init__(self, callback_id: CallbackId, output: str):
        super().__init__(callback_id=callback_id)
        self.output = output


class CallbackOutcomeFailure(CallbackOutcome):
    error: Optional[str]
    cause: Optional[str]

    def __init__(
        self, callback_id: CallbackId, error: Optional[str], cause: Optional[str]
    ):
        super().__init__(callback_id=callback_id)
        self.error = error
        self.cause = cause


class CallbackOutcomeTimedOut(CallbackOutcome):
    pass


class CallbackTimeoutError(TimeoutError):
    pass


class CallbackConsumerError(abc.ABC): ...


class CallbackConsumerTimeout(CallbackConsumerError):
    pass


class CallbackConsumerLeft(CallbackConsumerError):
    pass


class HeartbeatEndpoint:
    _mutex: Lock
    _next_heartbeat_event: Event
    _heartbeat_seconds: int

    def __init__(self, heartbeat_seconds: int):
        self._mutex = Lock()
        self._next_heartbeat_event = Event()
        self._heartbeat_seconds = heartbeat_seconds

    def clear_and_wait(self) -> bool:
        with self._mutex:
            if self._next_heartbeat_event.is_set():
                self._next_heartbeat_event.clear()
                return True
        return self._next_heartbeat_event.wait(timeout=self._heartbeat_seconds)

    def notify(self):
        with self._mutex:
            self._next_heartbeat_event.set()


class HeartbeatTimeoutError(TimeoutError):
    pass


class HeartbeatTimedOut(CallbackConsumerError):
    pass


class ActivityTaskStartOutcome:
    worker_name: Optional[str]

    def __init__(self, worker_name: Optional[str] = None):
        self.worker_name = worker_name


class ActivityTaskStartEndpoint:
    _next_activity_task_start_event: Event
    _outcome: Optional[ActivityTaskStartOutcome]

    def __init__(self):
        self._next_activity_task_start_event = Event()

    def wait(self, timeout_seconds: float) -> Optional[ActivityTaskStartOutcome]:
        self._next_activity_task_start_event.wait(timeout=timeout_seconds)
        return self._outcome

    def notify(self, activity_task: ActivityTaskStartOutcome) -> None:
        self._outcome = activity_task
        self._next_activity_task_start_event.set()


class CallbackEndpoint:
    callback_id: CallbackId
    _notify_event: Event
    _outcome: Optional[CallbackOutcome]
    consumer_error: Optional[CallbackConsumerError]
    _heartbeat_endpoint: Optional[HeartbeatEndpoint]

    def __init__(self, callback_id: CallbackId):
        self.callback_id = callback_id
        self._notify_event = Event()
        self._outcome = None
        self.consumer_error = None
        self._heartbeat_endpoint = None

    def setup_heartbeat_endpoint(self, heartbeat_seconds: int) -> HeartbeatEndpoint:
        self._heartbeat_endpoint = HeartbeatEndpoint(
            heartbeat_seconds=heartbeat_seconds
        )
        return self._heartbeat_endpoint

    def interrupt_all(self) -> None:
        # Interrupts all waiting processes on this endpoint.
        self._notify_event.set()
        heartbeat_endpoint = self._heartbeat_endpoint
        if heartbeat_endpoint is not None:
            heartbeat_endpoint.notify()

    def notify(self, outcome: CallbackOutcome):
        self._outcome = outcome
        self._notify_event.set()
        if self._heartbeat_endpoint:
            self._heartbeat_endpoint.notify()

    def notify_heartbeat(self) -> bool:
        if not self._heartbeat_endpoint:
            return False
        self._heartbeat_endpoint.notify()
        return True

    def wait(self, timeout: Optional[float] = None) -> Optional[CallbackOutcome]:
        self._notify_event.wait(timeout=timeout)
        return self._outcome

    def get_outcome(self) -> Optional[CallbackOutcome]:
        return self._outcome

    def report(self, consumer_error: CallbackConsumerError) -> None:
        self.consumer_error = consumer_error


class ActivityCallbackEndpoint(CallbackEndpoint):
    _activity_task_start_endpoint: ActivityTaskStartEndpoint
    _activity_input: str

    def __init__(self, callback_id: str, activity_input: str):
        super().__init__(callback_id=callback_id)
        self._activity_input = activity_input
        self._activity_task_start_endpoint = ActivityTaskStartEndpoint()

    def get_activity_input(self) -> str:
        return self._activity_input

    def get_activity_task_start_endpoint(self) -> ActivityTaskStartEndpoint:
        return self._activity_task_start_endpoint

    def notify_activity_task_start(self, worker_name: Optional[str]) -> None:
        self._activity_task_start_endpoint.notify(
            ActivityTaskStartOutcome(worker_name=worker_name)
        )


class CallbackNotifyConsumerError(RuntimeError):
    callback_consumer_error: CallbackConsumerError

    def __init__(self, callback_consumer_error: CallbackConsumerError):
        self.callback_consumer_error = callback_consumer_error


class CallbackOutcomeFailureError(RuntimeError):
    callback_outcome_failure: CallbackOutcomeFailure

    def __init__(self, callback_outcome_failure: CallbackOutcomeFailure):
        self.callback_outcome_failure = callback_outcome_failure


class CallbackPoolManager:
    _activity_store: Dict[CallbackId, Activity]
    _pool: Dict[CallbackId, CallbackEndpoint]

    def __init__(self, activity_store: Dict[Arn, Activity]):
        self._activity_store = activity_store
        self._pool = OrderedDict()

    def get(self, callback_id: CallbackId) -> Optional[CallbackEndpoint]:
        return self._pool.get(callback_id)

    def add(self, callback_id: CallbackId) -> CallbackEndpoint:
        if callback_id in self._pool:
            raise ValueError("Duplicate callback token id value.")
        callback_endpoint = CallbackEndpoint(callback_id=callback_id)
        self._pool[callback_id] = callback_endpoint
        return callback_endpoint

    def add_activity_task(
        self, callback_id: CallbackId, activity_arn: Arn, activity_input: str
    ) -> ActivityCallbackEndpoint:
        if callback_id in self._pool:
            raise ValueError("Duplicate callback token id value.")

        maybe_activity: Optional[Activity] = self._activity_store.get(activity_arn)
        if maybe_activity is None:
            raise ActivityDoesNotExist()

        maybe_activity.add_task(
            ActivityTask(task_token=callback_id, task_input=activity_input)
        )

        callback_endpoint = ActivityCallbackEndpoint(
            callback_id=callback_id, activity_input=activity_input
        )
        self._pool[callback_id] = callback_endpoint
        return callback_endpoint

    def generate(self) -> CallbackEndpoint:
        return self.add(long_uid())

    def notify(self, callback_id: CallbackId, outcome: CallbackOutcome) -> bool:
        callback_endpoint = self._pool.get(callback_id, None)
        if callback_endpoint is None:
            return False

        consumer_error: Optional[CallbackConsumerError] = (
            callback_endpoint.consumer_error
        )
        if consumer_error is not None:
            raise CallbackNotifyConsumerError(callback_consumer_error=consumer_error)

        callback_endpoint.notify(outcome=outcome)
        return True

    def heartbeat(self, callback_id: CallbackId) -> bool:
        callback_endpoint = self._pool.get(callback_id, None)
        if callback_endpoint is None:
            return False

        consumer_error: Optional[CallbackConsumerError] = (
            callback_endpoint.consumer_error
        )
        if consumer_error is not None:
            raise CallbackNotifyConsumerError(callback_consumer_error=consumer_error)

        return callback_endpoint.notify_heartbeat()



import moto.server as server

"""
Test the different server responses
"""


def test_elb_describe_instances():
    backend = server.create_backend_app("elb")
    test_client = backend.test_client()

    res = test_client.get("/?Action=DescribeLoadBalancers&Version=2015-12-01")

    assert b"DescribeLoadBalancersResponse" in res.data



"""qldb base URL and path."""

from .responses import QLDBResponse

url_bases = [
    r"https?://qldb\.(.+)\.amazonaws\.com",
]

url_paths = {
    "{0}/ledgers$": QLDBResponse.dispatch,
    "{0}/ledgers/(?P<ledger_name>[^/]+)$": QLDBResponse.dispatch,
    "{0}/tags/(?P<resource_arn>[^/]+)$": QLDBResponse.dispatch,
    "{0}/tags/(?P<resource_arn>[^/]+)/mock_ledger$": QLDBResponse.dispatch,
}



import xmltodict

import moto.server as server
from moto import mock_aws


@mock_aws
def test_cloudfront_list():
    backend = server.create_backend_app("cloudfront")
    test_client = backend.test_client()

    res = test_client.get("/2020-05-31/distribution")
    data = xmltodict.parse(res.data, dict_constructor=dict)
    assert "DistributionList" in data
    assert "Items" not in data["DistributionList"]



import json
import os
from functools import wraps
from uuid import uuid4

import boto3
import pytest
import requests
from botocore.exceptions import ClientError

from moto import mock_aws

from . import base_url, verify_execution_result


def aws_verified(create_table: bool = True):
    def inner(func):
        """
        Function that is verified to work against AWS.
        Can be run against AWS at any time by setting:
          MOTO_TEST_ALLOW_AWS_REQUEST=true

        If this environment variable is not set, the function runs in a `mock_aws` context.

        This decorator will:
          - Create an IAM-role that can be used by AWSLambda functions table
          - Run the test
          - Delete the role
        """

        @wraps(func)
        def pagination_wrapper():
            table_name = "table_" + str(uuid4())[0:6]

            allow_aws_request = (
                os.environ.get("MOTO_TEST_ALLOW_AWS_REQUEST", "false").lower() == "true"
            )

            if allow_aws_request:
                return create_table_and_test(table_name, sleep_time=10)
            else:
                with mock_aws():
                    requests.post(
                        f"http://{base_url}/moto-api/config",
                        json={"stepfunctions": {"execute_state_machine": True}},
                    )
                    resp = create_table_and_test(table_name, sleep_time=0)
                    requests.post(
                        f"http://{base_url}/moto-api/config",
                        json={"stepfunctions": {"execute_state_machine": False}},
                    )
                    return resp

        def create_table_and_test(table_name, sleep_time):
            if create_table:
                return _create_table_and_test(table_name, sleep_time)
            else:
                return func(table_name, sleep_time)

        def _create_table_and_test(table_name, sleep_time):
            client = boto3.client("dynamodb", region_name="us-east-1")

            client.create_table(
                TableName=table_name,
                KeySchema=[{"AttributeName": "id", "KeyType": "HASH"}],
                AttributeDefinitions=[{"AttributeName": "id", "AttributeType": "S"}],
                ProvisionedThroughput={"ReadCapacityUnits": 1, "WriteCapacityUnits": 5},
                Tags=[{"Key": "environment", "Value": "moto_tests"}],
            )
            waiter = client.get_waiter("table_exists")
            waiter.wait(TableName=table_name)
            try:
                resp = func(table_name, sleep_time)
            finally:
                ### CLEANUP ###
                client.delete_table(TableName=table_name)

            return resp

        return pagination_wrapper

    return inner


@aws_verified()
@pytest.mark.aws_verified
def test_state_machine_calling_dynamodb_put(table_name=None, sleep_time=0):
    dynamodb = boto3.client("dynamodb", "us-east-1")
    exec_input = {
        "TableName": table_name,
        "Item": {"data": {"S": "HelloWorld"}, "id": {"S": "id1"}},
    }

    expected_status = "SUCCEEDED"
    tmpl_name = "services/dynamodb_put_item"

    def _verify_result(client, execution, execution_arn):
        assert "stopDate" in execution
        assert json.loads(execution["input"]) == exec_input

        items = dynamodb.scan(TableName=table_name)["Items"]
        assert items == [exec_input["Item"]]

    verify_execution_result(
        _verify_result,
        expected_status,
        tmpl_name,
        exec_input=json.dumps(exec_input),
        sleep_time=sleep_time,
    )


@aws_verified()
@pytest.mark.aws_verified
def test_state_machine_calling_dynamodb_put_wait_for_invalid_task_token(
    table_name=None, sleep_time=0
):
    dynamodb = boto3.client("dynamodb", "us-east-1")
    exec_input = {
        "TableName": table_name,
        "Item": {"data": {"S": "HelloWorld"}, "id": {"S": "id1"}},
    }

    tmpl_name = "services/dynamodb_invalid_task_token"

    def _verify_result(client, execution, execution_arn):
        if execution["status"] == "RUNNING":
            items = dynamodb.scan(TableName=table_name)["Items"]
            if len(items) > 0:
                assert len(items) == 1
                assert items[0]["id"] == {"S": "1"}
                assert items[0]["StepFunctionTaskToken"] == {"S": "$$.Task.Token"}
                # Because the TaskToken is not returned, we can't signal to SFN to finish the execution
                # So let's just stop the execution manually
                client.stop_execution(
                    executionArn=execution_arn,
                    error="Bad Example - no TaskToken available to continue this execution",
                )
                return True

        return False

    verify_execution_result(
        _verify_result,
        expected_status=None,
        tmpl_name=tmpl_name,
        exec_input=json.dumps(exec_input),
        sleep_time=sleep_time,
    )


@aws_verified()
@pytest.mark.aws_verified
def test_state_machine_calling_dynamodb_put_wait_for_task_token(
    table_name=None, sleep_time=0
):
    dynamodb = boto3.client("dynamodb", "us-east-1")
    exec_input = {"TableName": table_name, "Item": {"id": {"S": "id1"}}}
    output = {"a": "b"}

    tmpl_name = "services/dynamodb_task_token"

    def _verify_result(client, execution, execution_arn):
        if execution["status"] == "RUNNING":
            items = dynamodb.scan(TableName=table_name)["Items"]
            if len(items) > 0:
                assert len(items) == 1
                assert items[0]["id"] == {"S": "1"}
                # Some random token
                assert len(items[0]["StepFunctionTaskToken"]["S"]) > 25
                token = items[0]["StepFunctionTaskToken"]["S"]
                client.send_task_success(taskToken=token, output=json.dumps(output))

        if execution["status"] == "SUCCEEDED":
            assert json.loads(execution["output"]) == output
            return True

        return False

    verify_execution_result(
        _verify_result,
        expected_status=None,
        tmpl_name=tmpl_name,
        exec_input=json.dumps(exec_input),
        sleep_time=sleep_time,
    )


@aws_verified()
@pytest.mark.aws_verified
def test_state_machine_calling_dynamodb_put_fail_task_token(
    table_name=None, sleep_time=0
):
    dynamodb = boto3.client("dynamodb", "us-east-1")
    exec_input = {"TableName": table_name, "Item": {"id": {"S": "id1"}}}

    tmpl_name = "services/dynamodb_task_token"

    def _verify_result(client, execution, execution_arn):
        if execution["status"] == "RUNNING":
            items = dynamodb.scan(TableName=table_name)["Items"]
            if len(items) > 0:
                assert len(items) == 1
                assert items[0]["id"] == {"S": "1"}
                # Some random token
                assert len(items[0]["StepFunctionTaskToken"]["S"]) > 25
                token = items[0]["StepFunctionTaskToken"]["S"]
                client.send_task_failure(taskToken=token, error="test error")

        if execution["status"] == "FAILED":
            assert execution["error"] == "test error"
            return True

        return False

    verify_execution_result(
        _verify_result,
        expected_status=None,
        tmpl_name=tmpl_name,
        exec_input=json.dumps(exec_input),
        sleep_time=sleep_time,
    )


@aws_verified()
@pytest.mark.aws_verified
def test_state_machine_calling_dynamodb_put_and_delete(table_name=None, sleep_time=0):
    dynamodb = boto3.client("dynamodb", "us-east-1")
    exec_input = {
        "TableName": table_name,
        "Item1": {"data": {"S": "HelloWorld"}, "id": {"S": "id1"}},
        "Item2": {"data": {"S": "HelloWorld"}, "id": {"S": "id2"}},
        "Key": {"id": {"S": "id1"}},
    }

    expected_status = "SUCCEEDED"
    tmpl_name = "services/dynamodb_put_delete_item"

    def _verify_result(client, execution, execution_arn):
        assert "stopDate" in execution
        assert json.loads(execution["input"]) == exec_input

        items = dynamodb.scan(TableName=table_name)["Items"]
        assert items == [exec_input["Item2"]]

    verify_execution_result(
        _verify_result,
        expected_status,
        tmpl_name,
        exec_input=json.dumps(exec_input),
        sleep_time=sleep_time,
    )


@aws_verified()
@pytest.mark.aws_verified
def test_send_task_failure_invalid_token(table_name=None, sleep_time=0):
    dynamodb = boto3.client("dynamodb", "us-east-1")
    exec_input = {"TableName": table_name, "Item": {"id": {"S": "id1"}}}

    tmpl_name = "services/dynamodb_task_token"

    def _verify_result(client, execution, execution_arn):
        if execution["status"] == "RUNNING":
            items = dynamodb.scan(TableName=table_name)["Items"]
            if len(items) > 0:
                # Execute
                with pytest.raises(ClientError) as exc:
                    client.send_task_failure(taskToken="bad_token", error="test error")

                # Verify
                assert exc.value.response["Error"]["Code"] == "InvalidToken"
                assert (
                    exc.value.response["Error"]["Message"]
                    == "Invalid Token: 'Invalid token'"
                )

                # Execute
                with pytest.raises(ClientError) as exc:
                    client.send_task_success(taskToken="bad_token", output="output")

                # Verify
                assert exc.value.response["Error"]["Code"] == "InvalidToken"
                assert (
                    exc.value.response["Error"]["Message"]
                    == "Invalid Token: 'Invalid token'"
                )

    verify_execution_result(
        _verify_result,
        expected_status=None,
        tmpl_name=tmpl_name,
        exec_input=json.dumps(exec_input),
        sleep_time=sleep_time,
    )


@aws_verified(create_table=False)
@pytest.mark.aws_verified
def test_zero_retry(table_name=None, sleep_time=0):
    exec_input = {"TableName": table_name, "Item": {"id": {"S": "id1"}}}

    tmpl_name = "services/dynamodb_zero_retry"

    def _verify_result(client, execution, execution_arn):
        if execution["status"] == "FAILED":
            assert execution["error"] == "DynamoDB.ResourceNotFoundException"
            assert (
                "Requested resource not found (Service: AmazonDynamoDBv2; Status Code: 400; Error Code: ResourceNotFoundException; Request ID:"
                in execution["cause"]
            )
            return True
        return False

    verify_execution_result(
        _verify_result,
        expected_status=None,
        tmpl_name=tmpl_name,
        exec_input=json.dumps(exec_input),
        sleep_time=sleep_time,
    )



import json
from typing import Any, Callable, Dict, Optional, Set

from botocore.exceptions import ClientError

from moto.stepfunctions.parser.api import (
    DescribeExecutionOutput,
    ExecutionStatus,
    HistoryEventType,
    TaskFailedEventDetails,
)
from moto.stepfunctions.parser.asl.component.common.error_name.custom_error_name import (
    CustomErrorName,
)
from moto.stepfunctions.parser.asl.component.common.error_name.failure_event import (
    FailureEvent,
    FailureEventException,
)
from moto.stepfunctions.parser.asl.component.common.error_name.states_error_name import (
    StatesErrorName,
)
from moto.stepfunctions.parser.asl.component.common.error_name.states_error_name_type import (
    StatesErrorNameType,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_task.credentials import (
    ComputedCredentials,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_task.service.resource import (
    ResourceCondition,
    ResourceRuntimePart,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_task.service.state_task_service_callback import (
    StateTaskServiceCallback,
)
from moto.stepfunctions.parser.asl.eval.environment import Environment
from moto.stepfunctions.parser.asl.eval.event.event_detail import EventDetails
from moto.stepfunctions.parser.asl.utils.boto_client import boto_client_for
from moto.stepfunctions.parser.asl.utils.encoding import to_json_str
from moto.utilities.collections import select_from_typed_dict

_SUPPORTED_INTEGRATION_PATTERNS: Set[ResourceCondition] = {
    ResourceCondition.WaitForTaskToken,
    ResourceCondition.Sync,
    ResourceCondition.Sync2,
}
_SUPPORTED_API_PARAM_BINDINGS: Dict[str, Set[str]] = {
    "startexecution": {"Input", "Name", "StateMachineArn"}
}


class StateTaskServiceSfn(StateTaskServiceCallback):
    def __init__(self):
        super().__init__(supported_integration_patterns=_SUPPORTED_INTEGRATION_PATTERNS)

    def _get_supported_parameters(self) -> Optional[Set[str]]:
        return _SUPPORTED_API_PARAM_BINDINGS.get(self.resource.api_action.lower())

    def _from_error(self, env: Environment, ex: Exception) -> FailureEvent:
        if isinstance(ex, ClientError):
            error_code = ex.response["Error"]["Code"]
            error_name: str = f"StepFunctions.{error_code}Exception"
            error_cause_details = [
                "Service: AWSStepFunctions",
                f"Status Code: {ex.response['ResponseMetadata']['HTTPStatusCode']}",
                f"Error Code: {error_code}",
                f"Request ID: {ex.response['ResponseMetadata']['RequestId']}",
                "Proxy: null",  # TODO: investigate this proxy value.
            ]
            if "HostId" in ex.response["ResponseMetadata"]:
                error_cause_details.append(
                    f'Extended Request ID: {ex.response["ResponseMetadata"]["HostId"]}'
                )
            error_cause: str = (
                f"{ex.response['Error']['Message']} ({'; '.join(error_cause_details)})"
            )
            return FailureEvent(
                env=env,
                error_name=CustomErrorName(error_name),
                event_type=HistoryEventType.TaskFailed,
                event_details=EventDetails(
                    taskFailedEventDetails=TaskFailedEventDetails(
                        error=error_name,
                        cause=error_cause,
                        resource=self._get_sfn_resource(),
                        resourceType=self._get_sfn_resource_type(),
                    )
                ),
            )
        return super()._from_error(env=env, ex=ex)

    def _normalise_parameters(
        self,
        parameters: dict,
        boto_service_name: Optional[str] = None,
        service_action_name: Optional[str] = None,
    ) -> None:
        if service_action_name is None:
            if self._get_boto_service_action() == "start_execution":
                optional_input = parameters.get("Input")
                if not isinstance(optional_input, str):
                    # AWS Sfn's documentation states:
                    # If you don't include any JSON input data, you still must include the two braces.
                    if optional_input is None:
                        optional_input = {}
                    parameters["Input"] = to_json_str(
                        optional_input, separators=(",", ":")
                    )
        super()._normalise_parameters(
            parameters=parameters,
            boto_service_name=boto_service_name,
            service_action_name=service_action_name,
        )

    def _build_sync_resolver(
        self,
        env: Environment,
        resource_runtime_part: ResourceRuntimePart,
        normalised_parameters: dict,
        task_credentials: ComputedCredentials,
    ) -> Callable[[], Optional[Any]]:
        sfn_client = boto_client_for(
            region=resource_runtime_part.region,
            account=resource_runtime_part.account,
            service="stepfunctions",
            credentials=task_credentials,
        )
        submission_output: dict = env.stack.pop()
        execution_arn: str = submission_output["ExecutionArn"]

        def _sync_resolver() -> Optional[Any]:
            describe_execution_output = sfn_client.describe_execution(
                executionArn=execution_arn
            )
            describe_execution_output: DescribeExecutionOutput = select_from_typed_dict(
                DescribeExecutionOutput, describe_execution_output
            )
            execution_status: ExecutionStatus = describe_execution_output["status"]

            if execution_status == ExecutionStatus.RUNNING:
                return None

            self._normalise_response(
                response=describe_execution_output,
                service_action_name="describe_execution",
            )
            if execution_status == ExecutionStatus.SUCCEEDED:
                return describe_execution_output
            else:
                raise FailureEventException(
                    FailureEvent(
                        env=env,
                        error_name=StatesErrorName(
                            typ=StatesErrorNameType.StatesTaskFailed
                        ),
                        event_type=HistoryEventType.TaskFailed,
                        event_details=EventDetails(
                            taskFailedEventDetails=TaskFailedEventDetails(
                                resource=self._get_sfn_resource(),
                                resourceType=self._get_sfn_resource_type(),
                                error=StatesErrorNameType.StatesTaskFailed.to_name(),
                                cause=to_json_str(describe_execution_output),
                            )
                        ),
                    )
                )

        return _sync_resolver

    @staticmethod
    def _sync2_api_output_of(typ: type, value: json) -> None:
        def _replace_with_json_if_str(key: str) -> None:
            inner_value = value.get(key)
            if isinstance(inner_value, str):
                value[key] = json.loads(inner_value)

        if typ == DescribeExecutionOutput:
            _replace_with_json_if_str("input")
            _replace_with_json_if_str("output")

    def _build_sync2_resolver(
        self,
        env: Environment,
        resource_runtime_part: ResourceRuntimePart,
        normalised_parameters: dict,
        task_credentials: ComputedCredentials,
    ) -> Callable[[], Optional[Any]]:
        sfn_client = boto_client_for(
            region=resource_runtime_part.region,
            account=resource_runtime_part.account,
            service="stepfunctions",
            credentials=task_credentials,
        )
        submission_output: dict = env.stack.pop()
        execution_arn: str = submission_output["ExecutionArn"]

        def _sync2_resolver() -> Optional[Any]:
            describe_execution_output = sfn_client.describe_execution(
                executionArn=execution_arn
            )
            describe_execution_output: DescribeExecutionOutput = select_from_typed_dict(
                DescribeExecutionOutput, describe_execution_output
            )
            execution_status: ExecutionStatus = describe_execution_output["status"]

            if execution_status == ExecutionStatus.RUNNING:
                return None

            self._sync2_api_output_of(
                typ=DescribeExecutionOutput, value=describe_execution_output
            )
            self._normalise_response(
                response=describe_execution_output,
                service_action_name="describe_execution",
            )
            if execution_status == ExecutionStatus.SUCCEEDED:
                return describe_execution_output
            else:
                raise FailureEventException(
                    FailureEvent(
                        env=env,
                        error_name=StatesErrorName(
                            typ=StatesErrorNameType.StatesTaskFailed
                        ),
                        event_type=HistoryEventType.TaskFailed,
                        event_details=EventDetails(
                            taskFailedEventDetails=TaskFailedEventDetails(
                                resource=self._get_sfn_resource(),
                                resourceType=self._get_sfn_resource_type(),
                                error=StatesErrorNameType.StatesTaskFailed.to_name(),
                                cause=to_json_str(describe_execution_output),
                            )
                        ),
                    )
                )

        return _sync2_resolver

    def _eval_service_task(
        self,
        env: Environment,
        resource_runtime_part: ResourceRuntimePart,
        normalised_parameters: dict,
        task_credentials: ComputedCredentials,
    ):
        service_name = self._get_boto_service_name()
        api_action = self._get_boto_service_action()
        sfn_client = boto_client_for(
            region=resource_runtime_part.region,
            account=resource_runtime_part.account,
            service=service_name,
            credentials=task_credentials,
        )
        response = getattr(sfn_client, api_action)(**normalised_parameters)
        response.pop("ResponseMetadata", None)
        env.stack.append(response)



"""TransferBackend class with methods for supported APIs."""

from typing import Any, Dict, List, Optional, Tuple

from moto.core.base_backend import BackendDict, BaseBackend
from moto.core.utils import unix_time
from moto.transfer.exceptions import PublicKeyNotFound, ServerNotFound, UserNotFound

from .types import (
    Server,
    ServerDomain,
    ServerEndpointType,
    ServerIdentityProviderType,
    ServerProtocols,
    User,
    UserHomeDirectoryType,
)


class TransferBackend(BaseBackend):
    """Implementation of Transfer APIs."""

    def __init__(self, region_name: str, account_id: str) -> None:
        super().__init__(region_name, account_id)
        self.servers: Dict[str, Server] = {}

    def create_server(
        self,
        certificate: Optional[str],
        domain: Optional[ServerDomain],
        endpoint_details: Optional[Dict[str, Any]],
        endpoint_type: Optional[ServerEndpointType],
        host_key: str,
        identity_provider_details: Optional[Dict[str, Any]],
        identity_provider_type: Optional[ServerIdentityProviderType],
        logging_role: Optional[str],
        post_authentication_login_banner: Optional[str],
        pre_authentication_login_banner: Optional[str],
        protocols: Optional[List[ServerProtocols]],
        protocol_details: Optional[Dict[str, Any]],
        security_policy_name: Optional[str],
        tags: Optional[List[Dict[str, str]]],
        workflow_details: Optional[Dict[str, Any]],
        structured_log_destinations: Optional[List[str]],
        s3_storage_options: Optional[Dict[str, Optional[str]]],
    ) -> str:
        server = Server(
            certificate=certificate,
            domain=domain,
            endpoint_type=endpoint_type,
            host_key_fingerprint=host_key,
            identity_provider_type=identity_provider_type,
            logging_role=logging_role,
            post_authentication_login_banner=post_authentication_login_banner,
            pre_authentication_login_banner=pre_authentication_login_banner,
            protocols=protocols,
            security_policy_name=security_policy_name,
            structured_log_destinations=structured_log_destinations,
            tags=(tags or []),
        )
        if endpoint_details is not None:
            endpoint_details = {
                "address_allocation_ids": endpoint_details.get("AddressAllocationIds"),
                "subnet_ids": endpoint_details.get("SubnetIds"),
                "vpc_endpoint_id": endpoint_details.get("VpcEndpointId"),
                "vpc_id": endpoint_details.get("VpcId"),
                "security_group_ids": endpoint_details.get("SecurityGroupIds"),
            }
            server.endpoint_details = endpoint_details
        if identity_provider_details is not None:
            identity_provider_details = {
                "url": identity_provider_details.get("Url"),
                "invocation_role": identity_provider_details.get("InvocationRole"),
                "directory_id": identity_provider_details.get("DirectoryId"),
                "function": identity_provider_details.get("Function"),
                "sftp_authentication_methods": identity_provider_details.get(
                    "SftpAuthenticationMethods"
                ),
            }
            server.identity_provider_details = identity_provider_details
        if protocol_details is not None:
            protocol_details = {
                "passive_ip": protocol_details.get("PassiveIp"),
                "tls_session_resumption_mode": protocol_details.get(
                    "TlsSessionResumptionMode"
                ),
                "set_stat_option": protocol_details.get("SetStatOption"),
                "as2_transports": protocol_details.get("As2Transports"),
            }
            server.protocol_details = protocol_details
        if s3_storage_options is not None:
            server.s3_storage_options = {
                "directory_listing_optimization": s3_storage_options.get(
                    "DirectoryListingOptimization"
                )
            }
        if workflow_details is not None:
            server.workflow_details = {
                "on_upload": [
                    {
                        "workflow_id": workflow.get("WorkflowId"),
                        "execution_role": workflow.get("ExecutionRole"),
                    }
                    for workflow in (workflow_details.get("OnUpload") or [])
                ],
                "on_partial_upload": [
                    {
                        "workflow_id": workflow.get("WorkflowId"),
                        "execution_role": workflow.get("ExecutionRole"),
                    }
                    for workflow in (workflow_details.get("OnPartialUpload") or [])
                ],
            }
        server_id = server.server_id
        self.servers[server_id] = server
        return server_id

    def describe_server(self, server_id: str) -> Server:
        if server_id not in self.servers:
            ServerNotFound(server_id=server_id)
        server = self.servers[server_id]
        return server

    def delete_server(self, server_id: str) -> None:
        if server_id not in self.servers:
            ServerNotFound(server_id=server_id)
        del self.servers[server_id]
        return

    def create_user(
        self,
        home_directory: Optional[str],
        home_directory_type: Optional[UserHomeDirectoryType],
        home_directory_mappings: Optional[List[Dict[str, Optional[str]]]],
        policy: Optional[str],
        posix_profile: Optional[Dict[str, Any]],
        role: str,
        server_id: str,
        ssh_public_key_body: Optional[str],
        tags: Optional[List[Dict[str, str]]],
        user_name: str,
    ) -> Tuple[str, str]:
        if server_id not in self.servers:
            ServerNotFound(server_id=server_id)
        user = User(
            home_directory=home_directory,
            home_directory_type=home_directory_type,
            policy=policy,
            role=role,
            tags=(tags or []),
            user_name=user_name,
        )
        if home_directory_mappings:
            for mapping in home_directory_mappings:
                user.home_directory_mappings.append(
                    {
                        "entry": mapping.get("Entry"),
                        "target": mapping.get("Target"),
                        "type": mapping.get("Type"),
                    }
                )
        if posix_profile is not None:
            posix_profile = {
                "gid": posix_profile.get("Gid"),
                "uid": posix_profile.get("Uid"),
                "secondary_gids": posix_profile.get("SecondaryGids"),
            }
            user.posix_profile = posix_profile
        if ssh_public_key_body is not None:
            now = unix_time()
            ssh_public_keys = [
                {
                    "date_imported": str(now),
                    "ssh_public_key_body": ssh_public_key_body,
                    "ssh_public_key_id": "mock_ssh_public_key_id_{ssh_public_key_body}_{now}",
                }
            ]
            user.ssh_public_keys = ssh_public_keys
        self.servers[server_id]._users.append(user)
        self.servers[server_id].user_count += 1
        return server_id, user_name

    def describe_user(self, server_id: str, user_name: str) -> Tuple[str, User]:
        if server_id not in self.servers:
            raise ServerNotFound(server_id=server_id)
        for user in self.servers[server_id]._users:
            if user.user_name == user_name:
                return server_id, user
        raise UserNotFound(user_name=user_name, server_id=server_id)

    def delete_user(self, server_id: str, user_name: str) -> None:
        if server_id not in self.servers:
            raise ServerNotFound(server_id=server_id)
        for i, user in enumerate(self.servers[server_id]._users):
            if user.user_name == user_name:
                del self.servers[server_id]._users[i]
                self.servers[server_id].user_count -= 1
                return
        raise UserNotFound(server_id=server_id, user_name=user_name)

    def import_ssh_public_key(
        self, server_id: str, ssh_public_key_body: str, user_name: str
    ) -> Tuple[str, str, str]:
        if server_id not in self.servers:
            raise ServerNotFound(server_id=server_id)
        for user in self.servers[server_id]._users:
            if user.user_name == user_name:
                date_imported = unix_time()
                ssh_public_key_id = (
                    f"{server_id}:{user_name}:public_key:{date_imported}"
                )
                key = {
                    "ssh_public_key_id": ssh_public_key_id,
                    "ssh_public_key_body": ssh_public_key_body,
                    "date_imported": str(date_imported),
                }
                user.ssh_public_keys.append(key)
                return server_id, ssh_public_key_id, user_name
        raise UserNotFound(user_name=user_name, server_id=server_id)

    def delete_ssh_public_key(
        self, server_id: str, ssh_public_key_id: str, user_name: str
    ) -> None:
        if server_id not in self.servers:
            raise ServerNotFound(server_id=server_id)
        for i, user in enumerate(self.servers[server_id]._users):
            if user.user_name == user_name:
                for j, key in enumerate(
                    self.servers[server_id]._users[i].ssh_public_keys
                ):
                    if key["ssh_public_key_id"] == ssh_public_key_id:
                        del user.ssh_public_keys[j]
                        return
                raise PublicKeyNotFound(
                    user_name=user_name,
                    server_id=server_id,
                    ssh_public_key_id=ssh_public_key_id,
                )
        raise UserNotFound(user_name=user_name, server_id=server_id)


transfer_backends = BackendDict(TransferBackend, "transfer")



from __future__ import annotations

import abc
from typing import List, Optional

from moto.stepfunctions.parser.api import HistoryEventType, TaskTimedOutEventDetails
from moto.stepfunctions.parser.asl.component.common.error_name.failure_event import (
    FailureEvent,
)
from moto.stepfunctions.parser.asl.component.common.error_name.states_error_name import (
    StatesErrorName,
)
from moto.stepfunctions.parser.asl.component.common.error_name.states_error_name_type import (
    StatesErrorNameType,
)
from moto.stepfunctions.parser.asl.component.common.parargs import Parargs
from moto.stepfunctions.parser.asl.component.state.exec.execute_state import (
    ExecutionState,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_task.credentials import (
    ComputedCredentials,
    Credentials,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_task.service.resource import (
    Resource,
)
from moto.stepfunctions.parser.asl.component.state.state_props import StateProps
from moto.stepfunctions.parser.asl.eval.environment import Environment
from moto.stepfunctions.parser.asl.eval.event.event_detail import EventDetails


class StateTask(ExecutionState, abc.ABC):
    resource: Resource
    parargs: Optional[Parargs]
    credentials: Optional[Credentials]

    def __init__(self):
        super(StateTask, self).__init__(
            state_entered_event_type=HistoryEventType.TaskStateEntered,
            state_exited_event_type=HistoryEventType.TaskStateExited,
        )

    def from_state_props(self, state_props: StateProps) -> None:
        super(StateTask, self).from_state_props(state_props)
        self.resource = state_props.get(Resource)
        self.parargs = state_props.get(Parargs)
        self.credentials = state_props.get(Credentials)

    def _get_supported_parameters(self) -> Optional[Set[str]]:  # noqa
        return None

    def _eval_parameters(self, env: Environment) -> dict:
        # Eval raw parameters.
        parameters = dict()
        if self.parargs is not None:
            self.parargs.eval(env=env)
            parameters = env.stack.pop()

        # Handle supported parameters.
        supported_parameters = self._get_supported_parameters()
        if supported_parameters:
            unsupported_parameters: List[str] = [
                parameter
                for parameter in parameters.keys()
                if parameter not in supported_parameters
            ]
            for unsupported_parameter in unsupported_parameters:
                parameters.pop(unsupported_parameter, None)

        return parameters

    def _eval_credentials(self, env: Environment) -> ComputedCredentials:
        if not self.credentials:
            task_credentials = dict()
        else:
            self.credentials.eval(env=env)
            task_credentials = env.stack.pop()
        return task_credentials

    def _get_timed_out_failure_event(self, env: Environment) -> FailureEvent:
        return FailureEvent(
            env=env,
            error_name=StatesErrorName(typ=StatesErrorNameType.StatesTimeout),
            event_type=HistoryEventType.TaskTimedOut,
            event_details=EventDetails(
                taskTimedOutEventDetails=TaskTimedOutEventDetails(
                    error=StatesErrorNameType.StatesTimeout.to_name(),
                )
            ),
        )

    def _from_error(self, env: Environment, ex: Exception) -> FailureEvent:
        if isinstance(ex, TimeoutError):
            return self._get_timed_out_failure_event(env)
        return super()._from_error(env=env, ex=ex)



from __future__ import annotations

import logging
import threading
from typing import Dict, Optional

from moto.stepfunctions.parser.api import (
    Arn,
    ExecutionStatus,
    InspectionLevel,
    StateMachineType,
    TestExecutionStatus,
    TestStateOutput,
    Timestamp,
)
from moto.stepfunctions.parser.asl.eval.evaluation_details import EvaluationDetails
from moto.stepfunctions.parser.asl.eval.program_state import (
    ProgramEnded,
    ProgramError,
    ProgramState,
)
from moto.stepfunctions.parser.asl.eval.test_state.program_state import (
    ProgramChoiceSelected,
)
from moto.stepfunctions.parser.asl.utils.encoding import to_json_str
from moto.stepfunctions.parser.backend.activity import Activity
from moto.stepfunctions.parser.backend.execution import (
    BaseExecutionWorkerCommunication,
    Execution,
)
from moto.stepfunctions.parser.backend.state_machine import StateMachineInstance
from moto.stepfunctions.parser.backend.test_state.execution_worker import (
    TestStateExecutionWorker,
)

LOG = logging.getLogger(__name__)


class TestStateExecution(Execution):
    exec_worker: Optional[TestStateExecutionWorker]
    next_state: Optional[str]

    class TestCaseExecutionWorkerCommunication(BaseExecutionWorkerCommunication):
        _execution: TestStateExecution

        def terminated(self) -> None:
            exit_program_state: ProgramState = (
                self.execution.exec_worker.env.program_state()
            )
            if isinstance(exit_program_state, ProgramChoiceSelected):
                self.execution.exec_status = ExecutionStatus.SUCCEEDED
                self.execution.output = (
                    self.execution.exec_worker.env.states.get_input()
                )
                self.execution.next_state = exit_program_state.next_state_name
            else:
                self._reflect_execution_status()

    def __init__(
        self,
        name: str,
        role_arn: Arn,
        exec_arn: Arn,
        account_id: str,
        region_name: str,
        state_machine: StateMachineInstance,
        start_date: Timestamp,
        activity_store: Dict[Arn, Activity],
        input_data: Optional[Dict] = None,
    ):
        super().__init__(
            name=name,
            sm_type=StateMachineType.STANDARD,
            role_arn=role_arn,
            exec_arn=exec_arn,
            account_id=account_id,
            region_name=region_name,
            state_machine=state_machine,
            start_date=start_date,
            activity_store=activity_store,
            input_data=input_data,
            cloud_watch_logging_session=None,
            trace_header=None,
        )
        self._execution_terminated_event = threading.Event()
        self.next_state = None

    def _get_start_execution_worker_comm(self) -> BaseExecutionWorkerCommunication:
        return self.TestCaseExecutionWorkerCommunication(self)

    def _get_start_execution_worker(self) -> TestStateExecutionWorker:
        return TestStateExecutionWorker(
            evaluation_details=EvaluationDetails(
                aws_execution_details=self._get_start_aws_execution_details(),
                execution_details=self.get_start_execution_details(),
                state_machine_details=self.get_start_state_machine_details(),
            ),
            exec_comm=self._get_start_execution_worker_comm(),
            cloud_watch_logging_session=self._cloud_watch_logging_session,
            activity_store=self._activity_store,
        )

    def publish_execution_status_change_event(self):
        # Do not publish execution status change events during test state execution.
        pass

    def to_test_state_output(
        self, inspection_level: InspectionLevel
    ) -> TestStateOutput:
        exit_program_state: ProgramState = self.exec_worker.env.program_state()
        if isinstance(exit_program_state, ProgramEnded):
            output_str = to_json_str(self.output)
            test_state_output = TestStateOutput(
                status=TestExecutionStatus.SUCCEEDED, output=output_str
            )
        elif isinstance(exit_program_state, ProgramError):
            test_state_output = TestStateOutput(
                status=TestExecutionStatus.FAILED,
                error=exit_program_state.error["error"],
                cause=exit_program_state.error["cause"],
            )
        elif isinstance(exit_program_state, ProgramChoiceSelected):
            output_str = to_json_str(self.output)
            test_state_output = TestStateOutput(
                status=TestExecutionStatus.SUCCEEDED,
                nextState=self.next_state,
                output=output_str,
            )
        else:
            # TODO: handle other statuses
            LOG.warning(
                "Unsupported StateMachine exit type for TestState '%s'",
                type(exit_program_state),
            )
            output_str = to_json_str(self.output)
            test_state_output = TestStateOutput(
                status=TestExecutionStatus.FAILED, output=output_str
            )

        if inspection_level == InspectionLevel.TRACE:
            test_state_output["inspectionData"] = self.exec_worker.env.inspection_data
        elif inspection_level == InspectionLevel.DEBUG:
            test_state_output["inspectionData"] = self.exec_worker.env.inspection_data

        return test_state_output



"""Exceptions raised by the s3tables service."""

from moto.core.exceptions import JsonRESTError


class BadRequestException(JsonRESTError):
    code = 400

    def __init__(self, message: str) -> None:
        super().__init__("BadRequestException", message)


class InvalidContinuationToken(BadRequestException):
    msg = "The continuation token is not valid."

    def __init__(self) -> None:
        super().__init__(self.msg)


class InvalidTableBucketName(BadRequestException):
    msg = "The specified bucket name is not valid."

    def __init__(self) -> None:
        super().__init__(self.msg)


class InvalidTableName(BadRequestException):
    template = "1 validation error detected: Value '%s' at 'name' failed to satisfy constraint: Member must satisfy regular expression pattern: [0-9a-z_]*"

    def __init__(self, name: str) -> None:
        super().__init__(self.template.format(name))


class InvalidNamespaceName(BadRequestException):
    msg = "The specified namespace name is not valid."

    def __init__(self) -> None:
        super().__init__(self.msg)


class InvalidMetadataLocation(BadRequestException):
    msg = "The specified metadata location is not valid."

    def __init__(self) -> None:
        super().__init__(self.msg)


class NothingToRename(BadRequestException):
    msg = "Neither a new namespace name nor a new table name is specified."

    def __init__(self) -> None:
        super().__init__(self.msg)


class NotFoundException(JsonRESTError):
    code = 404

    def __init__(self, message: str) -> None:
        super().__init__("NotFoundException", message)


class NamespaceDoesNotExist(NotFoundException):
    msg = "The specified namespace does not exist."

    def __init__(self) -> None:
        super().__init__(self.msg)


class DestinationNamespaceDoesNotExist(NotFoundException):
    msg = "The specified destination namespace does not exist."

    def __init__(self) -> None:
        super().__init__(self.msg)


class TableDoesNotExist(NotFoundException):
    msg = "The specified table does not exist."

    def __init__(self) -> None:
        super().__init__(self.msg)


class ConflictException(JsonRESTError):
    code = 409

    def __init__(self, message: str) -> None:
        super().__init__("ConflictException", message)


class VersionTokenMismatch(ConflictException):
    msg = "Provided version token does not match the table version token."

    def __init__(self) -> None:
        super().__init__(self.msg)


class TableAlreadyExists(ConflictException):
    msg = "A table with an identical name already exists in the namespace."

    def __init__(self) -> None:
        super().__init__(self.msg)



"""Unit tests for cloudhsmv2-supported APIs."""

import json
from datetime import datetime

import boto3

from moto import mock_aws

# See our Development Tips on writing tests for hints on how to write good tests:
# http://docs.getmoto.org/en/latest/docs/contributing/development_tips/tests.html


@mock_aws
def test_list_tags():
    client = boto3.client("cloudhsmv2", region_name="us-east-1")

    resource_id = "cluster-1234"
    client.tag_resource(
        ResourceId=resource_id,
        TagList=[
            {"Key": "Environment", "Value": "Production"},
            {"Key": "Project", "Value": "Security"},
        ],
    )

    response = client.list_tags(ResourceId=resource_id)
    assert len(response["TagList"]) == 2
    assert {"Key": "Environment", "Value": "Production"} in response["TagList"]
    assert {"Key": "Project", "Value": "Security"} in response["TagList"]
    assert "NextToken" not in response


@mock_aws
def test_tag_resource():
    client = boto3.client("cloudhsmv2", region_name="us-east-1")
    resource_id = "cluster-1234"

    response = client.tag_resource(
        ResourceId=resource_id,
        TagList=[
            {"Key": "Environment", "Value": "Production"},
            {"Key": "Project", "Value": "Security"},
        ],
    )

    tags = client.list_tags(ResourceId=resource_id)["TagList"]
    assert len(tags) == 2
    assert {"Key": "Environment", "Value": "Production"} in tags
    assert {"Key": "Project", "Value": "Security"} in tags

    response = client.tag_resource(
        ResourceId=resource_id,
        TagList=[{"Key": "Environment", "Value": "Development"}],
    )

    assert "ResponseMetadata" in response

    tags = client.list_tags(ResourceId=resource_id)["TagList"]
    assert len(tags) == 2
    assert {"Key": "Environment", "Value": "Development"} in tags
    assert {"Key": "Project", "Value": "Security"} in tags


@mock_aws
def test_list_tags_empty_resource():
    client = boto3.client("cloudhsmv2", region_name="us-east-1")

    response = client.list_tags(ResourceId="non-existent-resource")
    assert response["TagList"] == []
    assert "NextToken" not in response


@mock_aws
def test_untag_resource():
    client = boto3.client("cloudhsmv2", region_name="us-east-1")
    resource_id = "cluster-1234"

    client.tag_resource(
        ResourceId=resource_id,
        TagList=[
            {"Key": "Environment", "Value": "Production"},
            {"Key": "Project", "Value": "Security"},
            {"Key": "Team", "Value": "DevOps"},
        ],
    )

    initial_tags = client.list_tags(ResourceId=resource_id)["TagList"]
    assert len(initial_tags) == 3

    response = client.untag_resource(
        ResourceId=resource_id, TagKeyList=["Environment", "Team"]
    )
    assert "ResponseMetadata" in response

    remaining_tags = client.list_tags(ResourceId=resource_id)["TagList"]
    assert len(remaining_tags) == 1
    assert {"Key": "Project", "Value": "Security"} in remaining_tags


@mock_aws
def test_create_cluster():
    client = boto3.client("cloudhsmv2", region_name="us-east-1")

    response = client.create_cluster(
        BackupRetentionPolicy={"Type": "DAYS", "Value": "7"},
        HsmType="hsm1.medium",
        SubnetIds=["subnet-12345678"],
        TagList=[{"Key": "Environment", "Value": "Production"}],
    )

    cluster = response["Cluster"]
    assert cluster["BackupPolicy"] == "DEFAULT"
    assert cluster["BackupRetentionPolicy"] == {"Type": "DAYS", "Value": "7"}
    assert "ClusterId" in cluster
    assert isinstance(cluster["CreateTimestamp"], datetime)
    assert cluster["HsmType"] == "hsm1.medium"
    assert cluster["State"] == "ACTIVE"
    assert cluster["SubnetMapping"] == {"subnet-12345678": "us-east-1"}
    assert cluster["TagList"] == [{"Key": "Environment", "Value": "Production"}]
    assert "VpcId" in cluster

    clusters = client.describe_clusters()["Clusters"]
    assert len(clusters) == 1
    assert clusters[0]["ClusterId"] == cluster["ClusterId"]


@mock_aws
def test_delete_cluster():
    client = boto3.client("cloudhsmv2", region_name="us-east-1")

    response = client.create_cluster(
        HsmType="hsm1.medium",
        SubnetIds=["subnet-12345678"],
        NetworkType="IPV4",
        Mode="FIPS",
    )
    cluster_id = response["Cluster"]["ClusterId"]

    delete_response = client.delete_cluster(ClusterId=cluster_id)

    deleted_cluster = delete_response["Cluster"]
    assert deleted_cluster["ClusterId"] == cluster_id
    assert deleted_cluster["State"] == "DELETED"
    assert deleted_cluster["StateMessage"] == "Cluster deleted"

    clusters = client.describe_clusters()["Clusters"]
    assert len(clusters) == 0


@mock_aws
def test_describe_clusters_no_clusters():
    client = boto3.client("cloudhsmv2", region_name="us-east-1")
    response = client.describe_clusters()

    assert response["Clusters"] == []
    assert "NextToken" not in response


@mock_aws
def test_describe_clusters_with_filters():
    client = boto3.client("cloudhsmv2", region_name="us-east-1")

    cluster1 = client.create_cluster(
        HsmType="hsm1.medium",
        SubnetIds=["subnet-12345678"],
        NetworkType="IPV4",
        Mode="FIPS",
    )

    client.create_cluster(
        HsmType="hsm1.medium",
        SubnetIds=["subnet-87654321"],
        NetworkType="IPV4",
        Mode="FIPS",
    )

    response = client.describe_clusters(
        Filters={"clusterIds": [cluster1["Cluster"]["ClusterId"]]}
    )
    assert len(response["Clusters"]) == 1
    assert response["Clusters"][0]["ClusterId"] == cluster1["Cluster"]["ClusterId"]

    # Test filtering by state
    response = client.describe_clusters(Filters={"states": ["ACTIVE"]})
    assert len(response["Clusters"]) == 2


@mock_aws
def test_describe_clusters_pagination():
    client = boto3.client("cloudhsmv2", region_name="us-east-1")

    for i in range(3):
        client.create_cluster(
            HsmType="hsm1.medium",
            SubnetIds=[f"subnet-{i}"],
            NetworkType="IPV4",
            Mode="FIPS",
        )

    response = client.describe_clusters(MaxResults=2)
    assert len(response["Clusters"]) == 2
    assert "NextToken" in response

    # Get remaining clusters using NextToken
    response2 = client.describe_clusters(MaxResults=2, NextToken=response["NextToken"])
    assert len(response2["Clusters"]) == 1
    assert "NextToken" not in response2


@mock_aws
def test_get_resource_policy():
    client = boto3.client("cloudhsmv2", region_name="us-east-1")

    client.create_cluster(HsmType="hsm1.medium", SubnetIds=["subnet-12345678"])

    backup_response = client.describe_backups()
    backup_arn = backup_response["Backups"][0]["BackupArn"]

    policy = {
        "Version": "2012-10-17",
        "Statement": [
            {
                "Sid": "EnableSharing",
                "Effect": "Allow",
                "Principal": {"AWS": "arn:aws:iam::123456789012:root"},
                "Action": ["cloudhsmv2:DescribeBackups"],
                "Resource": backup_arn,
            }
        ],
    }

    client.put_resource_policy(ResourceArn=backup_arn, Policy=json.dumps(policy))
    response = client.get_resource_policy(ResourceArn=backup_arn)

    assert "Policy" in response
    assert json.loads(response["Policy"]) == policy


@mock_aws
def test_describe_backups():
    client = boto3.client("cloudhsmv2", region_name="us-east-1")

    cluster = client.create_cluster(
        HsmType="hsm1.medium",
        SubnetIds=["subnet-12345678"],
    )
    cluster_id = cluster["Cluster"]["ClusterId"]

    response = client.describe_backups()
    assert "Backups" in response
    assert len(response["Backups"]) == 1

    backup = response["Backups"][0]
    assert backup["ClusterId"] == cluster_id
    assert backup["HsmType"] == "hsm1.medium"
    assert backup["BackupState"] == "READY"

    filtered_response = client.describe_backups(Filters={"clusterIds": [cluster_id]})
    assert len(filtered_response["Backups"]) == 1
    assert filtered_response["Backups"][0]["ClusterId"] == cluster_id


@mock_aws
def test_put_resource_policy():
    client = boto3.client("cloudhsmv2", region_name="us-east-1")

    client.create_cluster(HsmType="hsm1.medium", SubnetIds=["subnet-12345678"])

    backup_response = client.describe_backups()
    backup_arn = backup_response["Backups"][0]["BackupArn"]

    policy = {
        "Version": "2012-10-17",
        "Statement": [
            {
                "Sid": "EnableSharing",
                "Effect": "Allow",
                "Principal": {"AWS": "arn:aws:iam::123456789012:root"},
                "Action": ["cloudhsmv2:DescribeBackups"],
                "Resource": backup_arn,
            }
        ],
    }

    response = client.put_resource_policy(
        ResourceArn=backup_arn, Policy=json.dumps(policy)
    )

    assert "ResourceArn" in response
    assert "Policy" in response
    assert response["ResourceArn"] == backup_arn
    assert json.loads(response["Policy"]) == policy


@mock_aws
def test_describe_backups_pagination():
    client = boto3.client("cloudhsmv2", region_name="us-east-1")

    # Create a cluster which will automatically create a backup
    for i in range(3):
        client.create_cluster(
            HsmType="hsm1.medium",
            SubnetIds=[f"subnet-{i}"],
            NetworkType="IPV4",
            Mode="FIPS",
        )

    assert len(client.describe_backups()["Backups"]) == 3

    # Test pagination with MaxResults
    response = client.describe_backups(MaxResults=2)
    assert len(response["Backups"]) == 2
    assert "NextToken" in response

    # Get remaining backups using NextToken
    response2 = client.describe_backups(MaxResults=2, NextToken=response["NextToken"])
    assert len(response2["Backups"]) == 1
    assert "NextToken" not in response2
